from copy import deepcopy
import numpy as np

import AvenirCommon.Util as acu

import SpectrumCommon.Const.PJ.PJNTags as pjmvt
import SpectrumCommon.Const.GB.GBConst as gbc

import SpectrumCommon.Const.IC.ICConst as icc
import SpectrumCommon.Const.IC.ICConstLink as iccl
import SpectrumCommon.Const.IC.ICTags as icmvt
import SpectrumCommon.Const.IC.ICModVarFields as icmvf
import SpectrumCommon.Const.IC.IC_TI_InterventionIDs as ictii
import SpectrumCommon.Const.IC.ICInterventionIDs as icii
import SpectrumCommon.Modvars.IC.ICModVarUtil as icmvu
import SpectrumCommon.Util.IC.ICUtilLink as icul

import SpectrumCommon.Const.IH.IHTags as ihmvt


# Notes:
#
#   1. The only tangible purpose of versions on MVs is to be able to update them appropriately for existing projections so those projections can run with the
#      current codebase. The only code that ever checks MV versions is in the XXUpdateModvars file. If you just added a new field, you could just check and see if
#      the field is in the MV and add it if it's not without changing the version on the MV. Similarly, if you added a new dimension, you could just extend the
#      dimension of a MV without creating a new version. Changing the version number of a MV does gives us more of a history to follow, but doesn't do more
#      for us programmatically. Also, if you change the version of a MV, all clients using that MV need to update their code to use the latest MV.
#      Changing the version of a MV is only really useful if the MV undergoes a major structural change that would make it difficult to
#      manipulate the existing version of that MV otherwise, or if we would like to apply a correction to all previous versions of a MV.
#
#   2. You cannot base logic on whether a current MV tag exists in the code below.  It will always exist, because before this gets called, all missing MVs
#      will get added to the projection.
#
#   3. To trigger this, create a projection, say on the client, save, and grab it's projection ID (in Chrome, F12 --> Application --> session storage).
#      Change GB_PM_Version in GBConst or UpdateModvars will get skipped.  Add proj ID to Postman environment. Call Login, then GetModvars to get the library ID and put that
#      into the Postman environment. Then call OpenProjection. Make sure to leave the projection on the client open  while doing this because the proj ID changes when
#      you open/close. Then call SaveProjection. You should now be able to call Calculate and see the modified MV there where it's used. Don't forget to change
#      the version number back after you finish testing and trash the test projection.
#
def IC_UpdateModvars(projection, created_projection):

    # Interventions

    # Previous versions of this updater seem to have left multiple versions of the intervention records (at which point there was just version 1 and 2), so
    # delete all but the most recent one if there is more than one.
    if (IC_INTERV_RECORDS_TAG_V1 in projection) and (IC_INTERV_RECORDS_TAG_V2 in projection):
        del projection[IC_INTERV_RECORDS_TAG_V1]

    if IC_INTERV_RECORDS_TAG_V1 in projection:
        _convert_interv_records_1(projection, IC_INTERV_RECORDS_TAG_V1)
        _convert_interv_records_2(projection, icmvt.IC_INTERV_RECORDS_TAG)
        _convert_interv_records_3(projection, icmvt.IC_INTERV_RECORDS_TAG)

    elif IC_INTERV_RECORDS_TAG_V2 in projection:
        _convert_interv_records_2(projection, IC_INTERV_RECORDS_TAG_V2)
        _convert_interv_records_3(projection, icmvt.IC_INTERV_RECORDS_TAG)

    elif IC_INTERV_RECORDS_TAG_V3 in projection:
        _convert_interv_records_3(projection, IC_INTERV_RECORDS_TAG_V3)

    if (projection[pjmvt.GB_IHTOnTag]) and (gbc.GB_CS in projection[pjmvt.PJN_ModulesTag]):
        # Added 2/14/2025 IHT now shows all LiST interventions except the FamPlan and HIV ones, so activate them. BF promo intervention was also non-impact.

        required_mod_vars = {
            iccl.IC_CS_IVInfoTag: projection[iccl.IC_CS_IVInfoTag],
            iccl.IC_CS_SelectedIVSetTag: projection[iccl.IC_CS_SelectedIVSetTag],
        }

        interv_to_exclude = []

        for mstID in acu.GBRange(iccl.CS_FirstFPMstID, iccl.CS_LastFPMstID):
            interv_to_exclude.append(mstID)

        interv_to_exclude.append(iccl.CS_FirstAMMstID)
        interv_to_exclude.append(iccl.CS_MstPercOnART)
        interv_to_exclude.append(int(iccl.CS_IV_PMTCT_MST_ID))
        interv_to_exclude.append(int(iccl.CS_IV_Cotrimoxazole_MST_ID))

        icul.IC_CS_set_interventions_active(required_mod_vars, interv_to_exclude)

        BF_promo_interv_rec = icmvu.get_IC_interv_rec(
            projection[icmvt.IC_INTERV_RECORDS_TAG], icii.IC_IV_ChangesInBF_MST_ID
        )
        if BF_promo_interv_rec != gbc.GB_NOT_FOUND:
            icmvu.set_IC_impact_interv(BF_promo_interv_rec, True)

        # Added 2/11/2025: IHT now shows part of LiST.  There's a checkbox for entering direct BF prevalence.  It should have defaulted on but defaulted off.  If off, there's also a special
        # mstID for BF promotion that should be in the selected intervention set and a MV used in calcs that needs to be set to promotion.

        # If direct entry of BF is not being used, assume promotion should be on.
        if iccl.IC_CS_MstDirectEntryOfBF not in projection[iccl.IC_CS_SelectedIVSetTag]:
            projection[iccl.IC_CS_SelectedIVSetTag].append(iccl.IC_CS_MstBFPromotion)
            projection[iccl.IC_CS_EnterBFImproveDataTag] = iccl.IC_CS_MstBFPromotionMode

    if IC_DRUG_SUPPLY_RECORDS_TAG_V1 in projection:
        _convert_drug_supply_records_1(projection, IC_DRUG_SUPPLY_RECORDS_TAG_V1)
        _convert_drug_supply_records_2(projection, icmvt.IC_DRUG_SUPPLY_RECORDS_TAG)

    elif IC_DRUG_SUPPLY_RECORDS_TAG_V2 in projection:
        _convert_drug_supply_records_2(projection, IC_DRUG_SUPPLY_RECORDS_TAG_V2)

    # Drug and supply treatment input records

    if IC_TREAT_INPUT_DRUG_SUPPLY_RECORDS_TAG_V1 in projection:
        # String constants were added for drugs and supplies, treatment input groups, and treatment inputs. Add them if they don't exist in those structures.
        TI_drug_supply_records = projection[IC_TREAT_INPUT_DRUG_SUPPLY_RECORDS_TAG_V1]

        for rec in TI_drug_supply_records:
            if icmvf.IC_TI_NOTE_STR_CONST not in rec:
                rec[icmvf.IC_TI_NOTE_STR_CONST] = ""

        projection[icmvt.IC_TREAT_INPUT_DRUG_SUPPLY_RECORDS_TAG] = deepcopy(TI_drug_supply_records)

    # Medical personnel treatment input records

    if IC_TREAT_INPUT_MED_PERSONNEL_RECORDS_TAG_V1 in projection:
        # String constants were added for drugs and supplies, treatment input groups, and treatment inputs. Add them if they don't exist in those structures.
        TI_med_personnel_records = projection[IC_TREAT_INPUT_MED_PERSONNEL_RECORDS_TAG_V1]

        for rec in TI_med_personnel_records:
            if icmvf.IC_TI_NOTE_STR_CONST not in rec:
                rec[icmvf.IC_TI_NOTE_STR_CONST] = ""

        projection[icmvt.IC_TREAT_INPUT_MED_PERSONNEL_RECORDS_TAG] = deepcopy(TI_med_personnel_records)

    # Outpatient visit/inpatient day treatment input records

    if IC_TREAT_INPUT_OUTPAT_INPAT_RECORDS_TAG_V1 in projection:
        # String constants were added for drugs and supplies, treatment input groups, and treatment inputs. Add them if they don't exist in those structures.
        TI_outpat_inpat_records = projection[IC_TREAT_INPUT_OUTPAT_INPAT_RECORDS_TAG_V1]

        for rec in TI_outpat_inpat_records:
            if icmvf.IC_TI_NOTE_STR_CONST not in rec:
                rec[icmvf.IC_TI_NOTE_STR_CONST] = ""

        projection[icmvt.IC_TREAT_INPUT_OUTPAT_INPAT_RECORDS_TAG] = deepcopy(TI_outpat_inpat_records)

    # Drug and supply group records

    if IC_TREAT_INPUT_GROUP_DRUG_SUPPLY_RECORDS_TAG_V1 in projection:
        # String constants were added for drugs and supplies, treatment input groups, and treatment inputs. Add them if they don't exist in those structures.
        TIG_drug_supply_records = projection[IC_TREAT_INPUT_GROUP_DRUG_SUPPLY_RECORDS_TAG_V1]

        for rec in TIG_drug_supply_records:
            if icmvf.IC_TIG_STR_CONST not in rec:
                rec[icmvf.IC_TIG_STR_CONST] = ""

        projection[icmvt.IC_TREAT_INPUT_GROUP_DRUG_SUPPLY_RECORDS_TAG] = deepcopy(TIG_drug_supply_records)

    # Medical personnel group records

    if IC_TREAT_INPUT_GROUP_MED_PERSONNEL_RECORDS_TAG_V1 in projection:
        # String constants were added for drugs and supplies, treatment input groups, and treatment inputs. Add them if they don't exist in those structures.
        TIG_med_personnel_records = projection[IC_TREAT_INPUT_GROUP_MED_PERSONNEL_RECORDS_TAG_V1]

        for rec in TIG_med_personnel_records:
            if icmvf.IC_TIG_STR_CONST not in rec:
                rec[icmvf.IC_TIG_STR_CONST] = ""

        projection[icmvt.IC_TREAT_INPUT_GROUP_MED_PERSONNEL_RECORDS_TAG] = deepcopy(TIG_med_personnel_records)

    # Outpatient visit/inpatient day group records

    if IC_TREAT_INPUT_GROUP_OUTPAT_INPAT_RECORDS_TAG_V1 in projection:
        # String constants were added for drugs and supplies, treatment input groups, and treatment inputs. Add them if they don't exist in those structures.
        TIG_outpat_inpat_records = projection[IC_TREAT_INPUT_GROUP_OUTPAT_INPAT_RECORDS_TAG_V1]

        for rec in TIG_outpat_inpat_records:
            if icmvf.IC_TIG_STR_CONST not in rec:
                rec[icmvf.IC_TIG_STR_CONST] = ""

        projection[icmvt.IC_TREAT_INPUT_GROUP_OUTPAT_INPAT_RECORDS_TAG] = deepcopy(TIG_outpat_inpat_records)

    if icmvt.IC_TARG_POP_RECORDS_TAG in projection:
        targ_pop_records_mv = projection[icmvt.IC_TARG_POP_RECORDS_TAG]

        TB_costing_on_mv = projection[pjmvt.GB_TBCostingOnTag]
        IHT_on_mv = projection[pjmvt.GB_IHTOnTag]

        found_355 = False
        found_356 = False
        found_357 = False

        # Target populations' meanings changed. Normally, we'd just make new target populations, but the ones we started with may not have
        # been available in the first place, so better to change them this time (and notice is short).
        for rec in targ_pop_records_mv:
            # Pulmonary and extrapulmonary TB, adults 15+: Number presumed rifampicin- and isoniazid-sensitive
            if rec["defTargPopMstID"] == 355:
                rec["name"] = (
                    "Pulmonary and extrapulmonary TB, adults 15+: Number presumed rifampicin- and isoniazid-sensitive"
                )
                rec["stringConstant"] = "GB_stPulmExtrapulmTBAdults15PPresRifampIsonSens"
                found_355 = True

            # Pulmonary and extrapulmonary TB, children 0-14: Number presumed rifampicin- and isoniazid-sensitive
            elif rec["defTargPopMstID"] == 356:
                rec["name"] = (
                    "Pulmonary and extrapulmonary TB, children 0-14: Number presumed rifampicin- and isoniazid-sensitive"
                )
                rec["stringConstant"] = "GB_stPulmExtrapulmTBChild0t14PresRifampIsonSens"
                found_356 = True

            # Pulmonary and extrapulmonary TB, children 0-14, without severe TB: Number presumed rifampicin- and isoniazid-sensitive
            elif rec["defTargPopMstID"] == 357:
                rec["name"] = (
                    "Pulmonary and extrapulmonary TB, children 0-14, without severe TB: Number presumed rifampicin- and isoniazid-sensitive"
                )
                rec["stringConstant"] = "GB_stPulmExtrapulmTBChild0t14NotSevPresRifampIsonSens"
                found_357 = True

        if TB_costing_on_mv or IHT_on_mv:
            # Forgot to add these new target populations for existing projections. There may be more elegant solutions (such as loading missing targ pops
            # from default database), but need to do the fastest one now, as we are in a time crunch.
            if not found_355:
                rec = {}
                rec["name"] = (
                    "Pulmonary and extrapulmonary TB, adults 15+: Number presumed rifampicin- and isoniazid-sensitive"
                )
                rec["stringConstant"] = "GB_stPulmExtrapulmTBAdults15PPresRifampIsonSens"
                rec["defTargPopMstID"] = 355
                rec["applicProgAreaIDs"] = ["22" if IHT_on_mv else "1"]
                rec["requiredModMstIDs"] = ["24"]
                targ_pop_records_mv.append(rec)

            if not found_356:
                rec = {}
                rec["name"] = (
                    "Pulmonary and extrapulmonary TB, children 0-14: Number presumed rifampicin- and isoniazid-sensitive"
                )
                rec["stringConstant"] = "GB_stPulmExtrapulmTBChild0t14PresRifampIsonSens"
                rec["defTargPopMstID"] = 356
                rec["applicProgAreaIDs"] = ["22" if IHT_on_mv else "1"]
                rec["requiredModMstIDs"] = ["24"]
                targ_pop_records_mv.append(rec)

            if not found_357:
                rec = {}
                rec["name"] = (
                    "Pulmonary and extrapulmonary TB, children 0-14, without severe TB: Number presumed rifampicin- and isoniazid-sensitive"
                )
                rec["stringConstant"] = "GB_stPulmExtrapulmTBChild0t14NotSevPresRifampIsonSens"
                rec["defTargPopMstID"] = 357
                rec["applicProgAreaIDs"] = ["22" if IHT_on_mv else "1"]
                rec["requiredModMstIDs"] = ["24"]
                targ_pop_records_mv.append(rec)

    if gbc.GB_TB in projection[pjmvt.PJN_ModulesTag]:
        # "Resistance testing: Bedaquiline, linezolid, levofloxacin, moxifloxacin" was not correctly marked as an
        # impact intervention in the intervention database. Add special case and flag appropriately to ensure
        # that coverage transfer works correctly for this intervention.
        interv_records_value = projection[icmvt.IC_INTERV_RECORDS_TAG]
        interv_rec = icmvu.get_IC_interv_rec(
            interv_records_value, ictii.IC_TI_IV_RESIST_TEST_BEDAQ_LINEZ_LEVO_MOXI_MST_ID
        )
        if interv_rec != gbc.GB_NOT_FOUND:
            special_cases = icmvu.get_IC_interv_special_cases(interv_rec)
            impact_interv = icmvu.get_IC_impact_interv(interv_rec)
            if icc.IC_TB_IMPACT not in special_cases:
                icmvu.add_IC_interv_special_case(interv_rec, icc.IC_TB_IMPACT)
            if not impact_interv:
                icmvu.set_IC_impact_interv(interv_rec, True)

    # delete tags
    for tag in IC_TagsForDeletion:
        if tag in projection:
            del projection[tag]

    return projection


# Old tags
IC_INTERV_RECORDS_TAG_V1 = "<IC_InterventionRecords_V1>"
IC_INTERV_RECORDS_TAG_V2 = "<IC_InterventionRecords_V2>"
IC_INTERV_RECORDS_TAG_V3 = "<IC_InterventionRecords_V3>"
IC_DRUG_SUPPLY_RECORDS_TAG_V1 = "<IC_DrugSupplyRecords_V1>"
IC_DRUG_SUPPLY_RECORDS_TAG_V2 = "<IC_DrugSupplyRecords_V2>"
IC_TREAT_INPUT_DRUG_SUPPLY_RECORDS_TAG_V1 = "<IC_TreatInputDrugSupplyRecords_V1>"
IC_TREAT_INPUT_MED_PERSONNEL_RECORDS_TAG_V1 = "<IC_TreatInputMedPersonnelRecords_V1>"
IC_TREAT_INPUT_OUTPAT_INPAT_RECORDS_TAG_V1 = "<IC_TreatInputOutpatInpatRecords_V1>"
IC_TREAT_INPUT_GROUP_DRUG_SUPPLY_RECORDS_TAG_V1 = "<IC_TreatInputGroupDrugSupplyRecords_V1>"
IC_TREAT_INPUT_GROUP_MED_PERSONNEL_RECORDS_TAG_V1 = "<IC_TreatInputGroupMedPersonnelRecords_V1>"
IC_TREAT_INPUT_GROUP_OUTPAT_INPAT_RECORDS_TAG_V1 = "<IC_TreatInputGroupOutpatInpatRecords_V1>"

IC_TagsForDeletion = [
    IC_INTERV_RECORDS_TAG_V1,
    IC_INTERV_RECORDS_TAG_V2,
    IC_INTERV_RECORDS_TAG_V3,
    IC_DRUG_SUPPLY_RECORDS_TAG_V1,
    IC_DRUG_SUPPLY_RECORDS_TAG_V2,
    IC_TREAT_INPUT_DRUG_SUPPLY_RECORDS_TAG_V1,
    IC_TREAT_INPUT_MED_PERSONNEL_RECORDS_TAG_V1,
    IC_TREAT_INPUT_OUTPAT_INPAT_RECORDS_TAG_V1,
    IC_TREAT_INPUT_GROUP_DRUG_SUPPLY_RECORDS_TAG_V1,
    IC_TREAT_INPUT_GROUP_MED_PERSONNEL_RECORDS_TAG_V1,
    IC_TREAT_INPUT_GROUP_OUTPAT_INPAT_RECORDS_TAG_V1,
]


def _convert_interv_records_1(projection, interv_records_tag):
    #  "TB preventive treatment, MDR/RR TB strain" is now a non-impact intervention.
    interv_records_value = projection[interv_records_tag]

    interv_rec = icmvu.get_IC_interv_rec(interv_records_value, ictii.IC_TI_IV_TBOralPrevMedWRifampDrugResist_MST_ID)

    if interv_rec != gbc.GB_NOT_FOUND:
        icmvu.set_IC_impact_interv(interv_rec, False)

    # Aa special case telling the client that the "TB preventive treatment, adults", "TB preventive treatment, children 0-14" interventions' coverages should be
    # locked did not exist. If this flag is not in the specialCases set for these interventions, add it.
    interv_rec = icmvu.get_IC_interv_rec(interv_records_value, ictii.IC_TI_IV_TB_ORAL_PREV_MED_ADULTS_15_PLUS_MST_ID)

    if interv_rec != gbc.GB_NOT_FOUND:
        icmvu.add_IC_interv_special_case(interv_rec, icc.IC_LOCK_COVERAGES)

    interv_rec = icmvu.get_IC_interv_rec(interv_records_value, ictii.IC_TI_IV_TB_ORAL_PREV_MED_CHILDREN_0_14_MST_ID)

    if interv_rec != gbc.GB_NOT_FOUND:
        icmvu.add_IC_interv_special_case(interv_rec, icc.IC_LOCK_COVERAGES)

    if interv_records_tag != icmvt.IC_INTERV_RECORDS_TAG:
        projection[icmvt.IC_INTERV_RECORDS_TAG] = deepcopy(interv_records_value)


def _convert_interv_records_2(projection, interv_records_tag):
    # Correction for issue where interventions were being created with a percent delivery channel distribution field that was the wrong length.
    interv_records_value = projection[interv_records_tag]
    deliv_chan_records_value = projection[ihmvt.IH_DELIV_CHAN_RECORDS_TAG_V1]

    num_deliv_chans_total = len(deliv_chan_records_value) + icc.IC_NUM_DELIV_CHANS

    for rec in interv_records_value:
        diff = num_deliv_chans_total - len(rec[icmvf.IC_INTERV_PERC_DELIV_THROUGH_CHAN])

        if num_deliv_chans_total > len(rec[icmvf.IC_INTERV_PERC_DELIV_THROUGH_CHAN]):
            for i in acu.GBRange(1, diff):
                rec[icmvf.IC_INTERV_PERC_DELIV_THROUGH_CHAN].append([0] * icc.IC_NUM_YEAR_ENDPOINTS)

        elif num_deliv_chans_total < len(rec[icmvf.IC_INTERV_PERC_DELIV_THROUGH_CHAN]):
            for i in acu.GBRange(1, diff):
                rec[icmvf.IC_INTERV_PERC_DELIV_THROUGH_CHAN].pop()

    if interv_records_tag != icmvt.IC_INTERV_RECORDS_TAG:
        projection[icmvt.IC_INTERV_RECORDS_TAG] = deepcopy(interv_records_value)


def _convert_drug_supply_records_1(projection, drug_supply_records_tag):
    # String constants were added for drugs and supplies, treatment input groups, and treatment inputs. Add them if they don't exist in those structures.
    drug_supply_records_value = projection[drug_supply_records_tag]

    for rec in drug_supply_records_value:
        if icmvf.IC_DRUG_SUPPLY_STR_CONST not in rec:
            rec[icmvf.IC_DRUG_SUPPLY_STR_CONST] = ""

    if drug_supply_records_tag != icmvt.IC_DRUG_SUPPLY_RECORDS_TAG:
        projection[icmvt.IC_DRUG_SUPPLY_RECORDS_TAG] = deepcopy(drug_supply_records_value)


def _convert_drug_supply_records_2(projection, drug_supply_records_tag):
    # For TB app, we added two new fields to the drug and supply records.
    drug_supply_records_value = projection[drug_supply_records_tag]

    first_year_mv_value = projection[pjmvt.PJN_FirstYearTag]
    final_year_mv_value = projection[pjmvt.PJN_FinalYearTag]

    num_years = final_year_mv_value - first_year_mv_value + 1

    if projection[pjmvt.GB_TBCostingOnTag]:
        for rec in drug_supply_records_value:
            if icmvf.IC_DRUG_SUPPLY_LOG_SUPP_CHAIN_COSTS_PERC not in rec:
                rec[icmvf.IC_DRUG_SUPPLY_LOG_SUPP_CHAIN_COSTS_PERC] = 0

            if icmvf.IC_DRUG_SUPPLY_WASTAGE_PERC not in rec:
                rec[icmvf.IC_DRUG_SUPPLY_WASTAGE_PERC] = 0

    # Several fields were added for Logistics. To make things easier, since Logistics can be turned on and off and the fields aren't huge,
    # just add them to all drug and supply records in all modes if they are not there.

    for rec in drug_supply_records_value:
        if icmvf.IC_DRUG_SUPPLY_PERC_WASTAGE not in rec:
            rec[icmvf.IC_DRUG_SUPPLY_PERC_WASTAGE] = 0

        if icmvf.IC_DRUG_SUPPLY_PERC_SAFETY_STOCK_NEEDED not in rec:
            rec[icmvf.IC_DRUG_SUPPLY_PERC_SAFETY_STOCK_NEEDED] = 0

        if icmvf.IC_DRUG_SUPPLY_SAFETY_STOCK_BASELINE not in rec:
            rec[icmvf.IC_DRUG_SUPPLY_SAFETY_STOCK_BASELINE] = 0

        if icmvf.IC_DRUG_SUPPLY_PERC_SAFETY_STOCK_REC not in rec:
            rec[icmvf.IC_DRUG_SUPPLY_PERC_SAFETY_STOCK_REC] = np.full((num_years), 0.0).tolist()

        if icmvf.IC_DRUG_SUPPLY_QUANTITY not in rec:
            rec[icmvf.IC_DRUG_SUPPLY_QUANTITY] = np.full((num_years), 0.0).tolist()

    if drug_supply_records_tag != icmvt.IC_DRUG_SUPPLY_RECORDS_TAG:
        projection[icmvt.IC_DRUG_SUPPLY_RECORDS_TAG] = deepcopy(drug_supply_records_value)


def _convert_interv_records_3(projection, interv_records_tag):
    # A special case telling the client that the "Testing of latent TB infection, other high-risk groups, adults 15+" interventions' coverage should be
    # locked did not exist. If this flag is not in the specialCases set for this intervention, add it.
    interv_records_value = projection[interv_records_tag]

    interv_rec = icmvu.get_IC_interv_rec(interv_records_value, ictii.IC_TI_IV_TestLatentTBOtherHRGsAdults15Plus_MST_ID)

    if interv_rec != gbc.GB_NOT_FOUND:
        icmvu.add_IC_interv_special_case(interv_rec, icc.IC_LOCK_COVERAGES)

    if interv_records_tag != icmvt.IC_INTERV_RECORDS_TAG:
        projection[icmvt.IC_INTERV_RECORDS_TAG] = deepcopy(interv_records_value)



from SpectrumCommon.Modvars.FP.FPModvarShapes import FP_modvar_shapes
from SpectrumCommon.Modvars.GB.GBDefs import gb_update_years

def FP_UpdateYears(projection, new_projection, params):
    gb_update_years(projection, params, FP_modvar_shapes)
    pass



from os import environ

from AvenirCommon.Database import GB_get_db_json, GB_file_exists

from AvenirCommon.Util import formatCountryFName, GBGetKeyValue
from SpectrumCommon.Const.DP import *

from SpectrumCommon.Modvars.DP.DPModvarDef import *


def init_modvars(countryID, first_year, final_year, extra):
    subnatCode = GBGetKeyValue(extra, 'subNatCode', 0)
    
    connection = environ[GB_SPECT_MOD_DATA_CONN_ENV]
    UPDCountryName = formatCountryFName(countryID, DPUPDVersion)#, subnatCode)
    if GB_file_exists(connection, 'demproj', 'UPD/' + UPDCountryName):
        UPDdb = GB_get_db_json(connection, 'demproj', 'UPD/' + UPDCountryName) 
    else:
        UPDdb = {}

    initialConditionsCountryName = formatCountryFName(countryID, DPInitialConditionsDBVersion)#, subnatCode)
    if GB_file_exists(connection, 'demproj', DPInitialConditionsDBSubDir + initialConditionsCountryName):
        InitialConditionsDB = GB_get_db_json(connection, 'demproj', DPInitialConditionsDBSubDir + initialConditionsCountryName) 
    else:
        InitialConditionsDB = {}
        
        
    DPGlobalData = GB_get_db_json(connection, 'demproj',  'DP_Global_' + DPDatabaseVersion + '.JSON')    
   
    projInfo = {
        'firstYr'           : first_year,
        'finalYr'           : final_year,
        'firstYrIdx'        : 0,
        'finalYrIdx'        : final_year - first_year,
        'subnatCode'        : GBGetKeyValue(extra, 'subNatCode', 0),
        'region'            : GBGetKeyValue(extra, 'region', DP_Asia),
    }
    survRate = DefaultUPDSurvRate_Init(projInfo, UPDdb)
    bigPop = BigPop_Init(projInfo, UPDdb, InitialConditionsDB)


    migrRate = MigrRate_Init(projInfo, UPDdb)    
    return {
        DP_DataSourceTag :  '' if UPDdb == {} else UPDdb['general']['Provider'] + ' ' + UPDdb['general']['Version'],
        DP_DPSourcesTag :  DPSources_Init(UPDdb),
        DP_DefaultUPDLETag :  DefaultUPDLE_Init(projInfo, UPDdb),
        DP_DefaultUPDIMRTag :  DefaultUPDIMR_Init(projInfo, UPDdb),
        DP_DefaultUPDCMRTag :  DefaultUPDCMR_Init(projInfo, UPDdb),
        DP_DefaultUPDSurvRateTag :  DefaultUPDSurvRate_Init(projInfo, UPDdb),
        DP_TFRTag :  TFR_Init(projInfo, UPDdb),
        DP_ASFRTag :  ASFR_Init(projInfo, UPDdb),
        DP_UNPopASFRTag :  UNPopASFR_Init(projInfo, UPDdb),
        DP_ASFRTablesTag :  ASFRTables_Init(DPGlobalData),
        DP_ASFRNumTag :  DP_ASFR_UNPop if 'pasfrs' in UPDdb else DP_ASFR_Average,
        DP_ASFRCustomFlagTag :  False,
        DP_SexBirthRatioTag :  SexBirthRatio_Init(projInfo, UPDdb),
        DP_LETag :  LE_Init(projInfo, UPDdb),
        DP_MigrRateTag :  migrRate,
        DP_MigrAgeDistTag :  MigrAgeDist_Init(projInfo, migrRate),
        DP_ModelLifeFileNameTag :  'DP_stInputted',
        DP_UseFYrSingleAgesTag :  GB_FP not in extra['modules'],
        DP_RegionalAdjustPopDataTag :  getList_sex_singleAge_year(projInfo['finalYrIdx']),
        DP_RegionalAdjustPopCBStateTag :  False,
        DP_RegionalAdjustPopFNameTag :  '',
        DP_RegionalAdjustPopFDateTag :  '',
        DP_CustomFileYearsTag :  CustomFilePYears_Init(),
        DP_CustomPopStopRescalingYearTag :  CustomPopStopRescalingYear_Init(projInfo),
        DP_LifeTableNumTag :  DP_UNPopLT, 
        DP_SurvRateTag : survRate,
        
        DP_BigPopTag :  bigPop,
        DP_FirstYearPop_Meta_Tag : np.array(bigPop)[:, :, 0].tolist(),
        DP_InitialConditionsTag : DP_InitialConditions_Init(projInfo, InitialConditionsDB),

        ## outputs ##
        DP_BirthsTag : getList_sex_5yrAge_year(projInfo['finalYrIdx']),
        DP_DeathsByAgeTag : getList_sex_singleAge_year(projInfo['finalYrIdx']),
        DP_IMRTag : getList_sex_year(projInfo['finalYrIdx']),
        DP_CMRTag : getList_sex_year(projInfo['finalYrIdx']),
        DP_AdjLETag : getList_sex_year(projInfo['finalYrIdx']), 

        # DP Temp Modvars

        DP_ResultAgeSpecificMortalityTag :  getList_sex_singleAge_year(projInfo['finalYrIdx']),
        DP_ResultDeathsByAgeLastBirthdayTag :  getList_sex_singleAge_year(projInfo['finalYrIdx']),
        DP_ResultCWRTag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultGRRTag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultMACTag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultNRRTag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultIMRTag :   getList_sex_year(projInfo['finalYrIdx']),
        DP_ResultU5MRTag :  getList_sex_year(projInfo['finalYrIdx']),
        DP_Result45q15Tag : getList_year(projInfo['finalYrIdx']),
        DP_ResultCBRTag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultCDRTag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultDoublingTimeTag : getList_year(projInfo['finalYrIdx']),
        DP_ResultGRTag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultRNITag :  getList_year(projInfo['finalYrIdx']),
        DP_ResultMedianAgeTag :  getList_sex_year(projInfo['finalYrIdx']),

        DP_UALowHighRatiosTag :  UALowHighRatios_Init(),
        DP_AIDSDeathsFYearUATag :  AIDSDeathsFYearUA_Init(),
        DP_UAHasRanTag :  UAHasRan_Init(),
        DP_UAInfoTag :  UAInfo_Init(),

    }
    
    



IC_TI_IV_BCG_Vacc_MST_ID = '49'
IC_TI_IV_ScreenProvInitPLHIV_GTE10YrsOnART_NotIll_MST_ID = '641'
IC_TI_IV_ScreenProvInitPLHIV_Children0t9_MST_ID = '551'
IC_TI_IV_ScreenProvInitPLHIV_GTE10YrsInpat_MST_ID = '550'
IC_TI_IV_ScreenProvInitPLHIV_Children0t9Inpat_MST_ID = '674'
IC_TI_IV_ScreenProvInitHHAdults15Plus_MST_ID = '555'
IC_TI_IV_ScreenProvInitHHChildren0t14_MST_ID = '552'
IC_TI_IV_ScreenProvInitHHChildren0t4_MST_ID = '553'
IC_TI_IV_ScreenProvInitHHChildren5t14_MST_ID = '554'
IC_TI_IV_ScreenProvInitPrisoners_MST_ID = '556'
IC_TI_IV_ScreenProvInitMiners_MST_ID = '557'
IC_TI_IV_ScreenProvInitFibroticLesions_MST_ID = '558'
IC_TI_IV_ScreenProvInitRiskFactorsGTE_P1PercTBPrev_MST_ID = '559'
IC_TI_IV_ScreenProvInitRiskFactorsLmtdHlthCare_MST_ID = '560'
IC_TI_IV_ScreenProvInitGenPop_MST_ID = '561'
IC_TI_IV_TBCaseFindPatInitAdults_MST_ID = '675'
IC_TI_IV_TBCaseFindPatInitChildren_MST_ID = '676'
IC_TI_IV_DiagPatInitPulmTBAdults_MST_ID = '574'
IC_TI_IV_DiagPatInitPulmTBChildren_MST_ID = '576'
IC_TI_IV_DiagPatInitExtraPulmTBAdults_MST_ID = '575'
IC_TI_IV_DiagPatInitExtraPulmTBChildren_MST_ID = '577'
IC_TI_IV_DiagProvInitPulmTB_PLHIV_Outpat_MST_ID = '580'
IC_TI_IV_DiagProvInitPulmTB_PLHIV_Inpat_MST_ID = '581'
IC_TI_IV_DiagProvInitPulmTB_PLHIV_ChildrenOutpat_MST_ID = '582'
IC_TI_IV_DiagProvInitPulmTB_PLHIV_ChildrenInpat_MST_ID = '583'
IC_TI_IV_DiagProvInitPulmTB_HHChildren_MST_ID = '579'
IC_TI_IV_DiagProvInitPulmTBAllHRGsAdults_MST_ID = '578'
IC_TI_IV_RifampIsonResistTest_MST_ID = '584'
IC_TI_IV_IsonResistTestRifampSensCases_MST_ID = '585'
IC_TI_IV_FluoroqResistTestRifampResistCases_MST_ID = '586'
IC_TI_IV_RESIST_TEST_BEDAQ_LINEZ_LEVO_MOXI_MST_ID = '649'
IC_TI_IV_TreatDrugSuscTBRifampIsonSensAdults_MST_ID = '597'
IC_TI_IV_TreatDrugSuscTBRifampIsonSensChildren0t14_MST_ID = '598'
IC_TI_IV_TreatDrugSuscTBRifampIsonSensChildren3MoTo16_MST_ID = '599'
IC_TI_IV_TreatTBMeningPericardAdults_MST_ID = '600'
IC_TI_IV_TreatDrugSuscTBMeningPericardChildren_MST_ID = '601'
IC_TI_IV_NonMedTreatSuppPatFirstLine_MST_ID = '677'
IC_TI_IV_TreatDrugResistTBRifampSensIsonResistAdults_MST_ID = '602'
IC_TI_IV_TreatDrugResistTBRifampSensIsonResistChildren_MST_ID = '603'
IC_TI_IV_BpalMOralBedaqRegDrugResistTBAdults_MST_ID = '604'
IC_TI_IV_OralBedaqRegDrugResistTBChildren_MST_ID = '605'
IC_TI_IV_BPalRegDrugResistTBAdults_MST_ID = '606'
IC_TI_IV_LongerRegDrugResistTBTreatAdults_MST_ID = '607'
IC_TI_IV_LongerRegDrugResistTBTreatChildren_MST_ID = '608'
IC_TI_IV_TreatDrugResistTB_XDR_Adults_MST_ID = '609'
IC_TI_IV_TreatDrugResistTB_XDR_Children_MST_ID = '610'
IC_TI_IV_NonMedTreatSuppPatSecLine_MST_ID = '678'
IC_TI_IV_TestLatentTB_HHContactAdults15Plus_MST_ID = '563'
IC_TI_IV_TestLatentTB_HHContactChildren5t14_MST_ID = '562'
IC_TI_IV_TestLatentTB_HHContactChildren0t4_MST_ID = '564'
IC_TI_IV_TestLatentTBOtherHRGsAdults15Plus_MST_ID = '565'
IC_TI_IV_TB_ORAL_PREV_MED_ADULTS_15_PLUS_MST_ID = '650'
IC_TI_IV_TB_ORAL_PREV_MED_CHILDREN_0_14_MST_ID = '567'
IC_TI_IV_TBOralPrevMedWRifampDrugResist_MST_ID = '569'
IC_TI_IV_LungResectionSurgery_MST_ID = '673'
IC_TI_IV_PalliativeCare_MST_ID = '625'



IH_VERSION_TAG_V1                                  = '<OH_Version_V1>'
IH_HLTH_SERV_ANAL_MODE_MST_ID_TAG_V1               = '<OH_HealthServiceAnalysisModeMstID_V1>'
IH_PROG_AREA_DELIV_CHAN_FILTER_MST_ID_TAG_V1       = '<OH_ProgrammeAreaDeliveryChannelFilterMstID_V1>'
IH_MAX_CUST_INTERV_MST_ID_TAG_V1                   = '<OH_MaxCustomInterventionMstID_V1>'     
IH_MOH_BUDGET_TAG_V1                               = '<OH_MOHBudget_V1>'               
IH_DATA_CHANGED_TAG_V1                             = '<OH_DataChanged_V1>'
IH_PROJ_VALID_TAG_V1                               = '<OH_ProjectionValid_V1>'
IH_USING_ADOL_STRATA_MODE_TAG_V1                   = '<OH_UsingAdolescentStrataMode_V1>'
IH_USING_DIS_SPEC_COST_MODE_TAG_V1                 = '<OH_UsingDiseaseSpecificCostingMode_V1>'
IH_USING_PHC_MODE_TAG_V1                           = '<OH_UsingPrimaryHealthCareMode_V1>'
IH_USING_DEF_MED_STAFF_TYPES_TAG_V1                = '<OH_UsingDefaultMedicalStaffTypes_V1>' 
IH_USING_DEF_FAC_TYPES_TAG_V1                      = '<OH_UsingDefaultFacilityTypes_V1>' 
IH_USING_MOH_BUDGET_RES_TAG_V1                     = '<OH_UsingMOHBudgetInResults_V1>' 
IH_USING_HPV_PROG_AREA_TAG_V1                      = '<OH_UsingHPVProgrammeArea_V1>'
IH_USING_WORK_HLTH_PROG_AREA_TAG_V1                = '<OH_UsingWorkersHealthProgrammeArea_V1>'
IH_USING_TP_FROM_TIME_TAG_V1                       = '<OH_UsingTargetPopulationsFromTIME_V1>'
IH_USING_TB_IMPACT_CALCS_TP_TAG_V1                 = '<OH_UsingTBImpactCalcsForTargetPopulations_V1>'
IH_USING_TB_DAT_TP_TAG_V1                          = '<OH_UsingTB_DAT_TargetPopulations_V1>'
IH_USING_LOCAL_CURR_TAG_V1                         = '<OH_UsingLocalCurrency_V1>'
IH_FILE_NAME_TAG_V1                                = '<OH_FileName_V1>'                        
IH_PROJ_VALID_DATE_TAG_V1                          = '<OH_ProjectionValidDate_V1>'                      
IH_PROG_AREA_RECORDS_TAG_V1                        = '<OH_ProgrammeAreaRecords_V1>'            
IH_DELIV_CHAN_RECORDS_TAG_V1                       = '<OH_DeliveryChannelRecords_V1>'         
IH_NAT_PROG_AREA_RECORDS_TAG_V1                    = '<OH_NationalProgrammeRecords_V1>'     
IH_AUDIT_RECORDS_TAG_V1                            = '<OH_AuditRecords_V1>'         
IH_HEALTH_SYS_ADMIN_UNIT_RECORDS_TAG               = '<OH_HealthSysAdminUnitRecords_V1>'
IH_LOCKED_AREA_RECORDS_TAG_V1                      = '<OH_LockedAreaRecords_V1>' 
IH_HPV_PARAM_RECORDS_TAG_V1                        = '<OH_HPV_ParamRecords_V1>'
IH_COLD_CHAIN_TAG_V1                               = '<OH_ColdChain_V1>'
IH_PERC_PROG_MGMT_COSTS_PHC_VIS_TAG_V1             = '<OH_PercProgOrMgmtCostsPHC_Visited_V1>'    
IH_PERC_PROG_MGMT_COSTS_PHC_PROG_AREA_TAG_V1       = '<OH_PercProgOrMgmtCostsPHC_ProgArea_V1>'
IH_PERC_PROG_MGMT_COSTS_PHC_HLTH_SYS_TAG_V1        = '<OH_PercProgOrMgmtCostsPHC_HealthSys_V1>'
IH_PERC_PROG_MGMT_COSTS_PHC_DELIV_CHAN_TAG_V1      = '<OH_PercProgOrMgmtCostsPHC_DelivChan_V1>'
IH_PERC_PROG_MGMT_COSTS_PHC_NAT_PROG_TAG_V1        = '<OH_PercProgOrMgmtCostsPHC_NatProg_V1>'
IH_PERC_PROG_MGMT_COSTS_PHC_DELIV_CHAN_PACK_TAG_V1 = '<OH_PercProgOrMgmtCostsPHC_DelivChanPack_V1>'
IH_HR_FACTOR_PHC_TAG_V1                            = '<OH_HRFactorPHC_V1>'    
IH_PERC_FAC_TYPE_PHC_TAG_V1                        = '<OH_PercFacTypePHC_V1>'     
IH_LG_PERC_DRUG_SUPP_COSTS_PHC_TAG_V1              = '<OH_LGPercDrugSuppCostsPHC_V1>'   
IH_SHOW_INACTIVE_PROG_AREAS_EXIST_INTERV_V1        = '<OH_ShowInactiveProgAreasExistIntervs_V1>'
IH_SHOW_INACTIVE_PROG_AREAS_EXIST_DRUG_SUPP_V1     = '<OH_ShowInactiveProgAreasExistDrugsSupplies_V1>'
IH_ENTER_UNIT_COSTS_ALL_YEARS_DRUG_SUPP_V1         = '<OH_EnterUnitCostsAllYearsDrugsSupplies_V1>'

# TB Costing

IH_PROG_AREA_RECORDS_IHT_ID_INFO_ONLY_TAG_V1       = '<OH_ProgrammeAreaRecordsIHT_IDInfoOnly_V1>'

# Results 

IH_RES_STAFF_CAP_UTILIZATION_STATIC_TAG_V1          = '<OH_ResStaffCapUtilizationStatic_V1>' 
IH_RES_STAFF_CAP_UTILIZATION_DYNAMIC_TAG_V1         = '<OH_ResStaffCapUtilizationDynamic_V1>' 



import numpy as np
from loguru import logger



def convert_to_numpy(projection):
    _dtype = np.float64
    dt = np.dtype(np.float64)  

    for tag in projection:
        if tag[0] == '<':
            mod_str = tag[1:3]
                
            if type(projection[tag]) == list:
                try:
                    if len(projection[tag])>0 and ((type(projection[tag][0])==dict) or(type(projection[tag][0])==str)) :
                        projection[tag] = np.array(projection[tag], order='C')#, dtype = dt)
                    else:
                        projection[tag] = np.array(projection[tag], order='C', dtype= dt )#, dtype = dt)
                except:
                    logger.warning(f"Could not convert {tag} to numpy array")
                    pass


def convert_to_list(projection):

    numpy_type = type(np.array([]))
    for tag in projection:
        if type(projection[tag]) == numpy_type:
            projection[tag] = np.nan_to_num(projection[tag], copy=True, nan=0.0, posinf=None, neginf=None)
            projection[tag] = projection[tag].tolist()



    
def remove_nonmodvars(projection):
    
    # Using a list comprehension to make a list of the keys to be deleted
    delete = [tag for tag in projection if tag[0] != '<']
    # delete the non tags key/s
    for key in delete:
        del projection[key]



import numpy as np
import math
from scipy.optimize import show_options

#% ======================================
#% ========= begin subfunctions =========
#% ======================================
def intrafun(x,params): #function fval = intrafun(x,params)
    #% transform variables, then call original function

    #% transform
    xtrans = xtransform(x,params)

    #% and call fun
    fval = params['fun'](np.reshape(xtrans,params['xsize']),params['args']) #fval = feval(params.fun,reshape(xtrans,params.xsize),params.args{:});

    #end % sub function intrafun end

    #% ======================================
    return (fval)

def xtransform(x,params): #function xtrans = xtransform(x,params)
    #% converts unconstrained variables into their original domains

    xtrans = np.zeros(params['xsize']) #xtrans = zeros(params.xsize);
    #% k allows some variables to be fixed, thus dropped from the
    #% optimization.
    k=1 #k=1;
    for i in range(params['n']): #i = 1:params.n
        match params['BoundClass'][i]: #switch params.BoundClass(i)
            case 1:
                #% lower bound only
                xtrans[i] = params['LB'][i] + x[k]*x[k] #xtrans(i) = params.LB(i) + x(k).^2;
      
                k=k+1 #k=k+1;
            case 2:
                #% upper bound only
                xtrans[i] = params['UB'][i] - x[k]*x[k] #xtrans(i) = params.UB(i) - x(k).^2;
      
                k=k+1 #k=k+1;
            case 3:
                #% lower and upper bounds
                xtrans[i] = (np.sin(x[k])+1)/2 #xtrans(i) = (sin(x(k))+1)/2;
                xtrans[i] = xtrans[i]*(params['UB'][i] - params['LB'][i]) + params['LB'][i]#xtrans(i) = xtrans(i)*(params.UB(i) - params.LB(i)) + params.LB(i);
                #% just in case of any floating point problems
                xtrans[i] = max(params['LB'][i],min(params['UB'][i],xtrans[i]))#xtrans(i) = max(params.LB(i),min(params.UB(i),xtrans(i)));
      
                k = k+1
            case 4:
                #% fixed variable, bounds are equal, set it at either bound
                xtrans[i] = params['LB'][i]
            case 0:
                #% unconstrained variable.
                xtrans[i] = x[k] #xtrans(i) = x(k);
      
                k=k+1
            #end
        #end
    return(xtrans)
    #end % sub function xtransform end

def fminsearchbnd(fun,x0,LB,UB,options,varargin): #function [x,fval,exitflag,output] = fminsearchbnd(fun,x0,LB,UB,options,varargin)
    #% FMINSEARCHBND: FMINSEARCH, but with bound constraints by transformation
    #% usage: x=FMINSEARCHBND(fun,x0)
    #% usage: x=FMINSEARCHBND(fun,x0,LB)
    #% usage: x=FMINSEARCHBND(fun,x0,LB,UB)
    #% usage: x=FMINSEARCHBND(fun,x0,LB,UB,options)
    #% usage: x=FMINSEARCHBND(fun,x0,LB,UB,options,p1,p2,...)
    #% usage: [x,fval,exitflag,output]=FMINSEARCHBND(fun,x0,...)
    #% 
    #% arguments:
    #%  fun, x0, options - see the help for FMINSEARCH
    #%
    #%  LB - lower bound vector or array, must be the same size as x0
    #%
    #%       If no lower bounds exist for one of the variables, then
    #%       supply -inf for that variable.
    #%
    #%       If no lower bounds at all, then LB may be left empty.
    #%
    #%       Variables may be fixed in value by setting the corresponding
    #%       lower and upper bounds to exactly the same value.
    #%
    #%  UB - upper bound vector or array, must be the same size as x0
    #%
    #%       If no upper bounds exist for one of the variables, then
    #%       supply +inf for that variable.
    #%
    #%       If no upper bounds at all, then UB may be left empty.
    #%
    #%       Variables may be fixed in value by setting the corresponding
    #%       lower and upper bounds to exactly the same value.
    #%
    #% Notes:
    #%
    #%  If options is supplied, then TolX will apply to the transformed
    #%  variables. All other FMINSEARCH parameters should be unaffected.
    #%
    #%  Variables which are constrained by both a lower and an upper
    #%  bound will use a sin transformation. Those constrained by
    #%  only a lower or an upper bound will use a quadratic
    #%  transformation, and unconstrained variables will be left alone.
    #%
    #%  Variables may be fixed by setting their respective bounds equal.
    #%  In this case, the problem will be reduced in size for FMINSEARCH.
    #%
    #%  The bounds are inclusive inequalities, which admit the
    #%  boundary values themselves, but will not permit ANY function
    #%  evaluations outside the bounds. These constraints are strictly
    #%  followed.
    #%
    #%  If your problem has an EXCLUSIVE (strict) constraint which will
    #%  not admit evaluation at the bound itself, then you must provide
    #%  a slightly offset bound. An example of this is a function which
    #%  contains the log of one of its parameters. If you constrain the
    #%  variable to have a lower bound of zero, then FMINSEARCHBND may
    #%  try to evaluate the function exactly at zero.
    #%
    #%
    #% Example usage:
    #% rosen = @(x) (1-x(1)).^2 + 105*(x(2)-x(1).^2).^2;
    #%
    #% fminsearch(rosen,[3 3])     % unconstrained
    #% ans =
    #%    1.0000    1.0000
    #%
    #% fminsearchbnd(rosen,[3 3],[2 2],[])     % constrained
    #% ans =
    #%    2.0000    4.0000
    #%
    #% See test_main.m for other examples of use.
    #%
    #%
    #% See also: fminsearch, fminspleas
    #%
    #%
    #% Author: John D'Errico
    #% E-mail: woodchips@rochester.rr.com
    #% Release: 4
    #% Release date: 7/23/06

    #% size checks
    xsize = np.size(x0)
    x0 = x0[:]
    n = len(x0)

    if (nargin<3) or len(LB)==0: #(nargin<3) || isempty(LB)
        LB = np.tile(-math.inf,n,1) #LB = repmat(-inf,n,1)
    else:
        LB = LB[:]
    #end
    if (nargin<4) or len(UB)==0: #(nargin<4) || isempty(UB)
       UB = np.tile(math.inf,n,1)
    else:
      UB = UB[:]
    #end

    if (n!=len(LB)) or (n !=len(UB)):
      raise ValueError("x0 is incompatible in size with either LB or UB.")
    #end

    #% set default options if necessary
    if (nargin<5) or len(options)==0:
        show_options(solver="minimize") #options = optimset('fminsearch')
    #end

    #% stuff into a struct to pass around
    params = []    
    params['args'] = varargin
    params['LB'] = LB #params.LB = LB;
    params['UB'] = UB #params.UB = UB;
    params['fun'] = fun #params.fun = fun;
    params['n'] = n#params.n = n;
    #% note that the number of parameters may actually vary if 
    #% a user has chosen to fix one or more parameters
    params['xsize'] = xsize #params.xsize = xsize;
    params['OutputFcn'] = []#params.OutputFcn = [];

    #% 0 --> unconstrained variable
    #% 1 --> lower bound only
    #% 2 --> upper bound only
    #% 3 --> dual finite bounds
    #% 4 --> fixed variable
    params['BoundClass'] = np.zeros[n,1] #params.BoundClass = zeros(n,1);
    for i in range(n): #i=1:n
        k = np.isfinite(LB[i]) + 2*np.isfinite(UB[i]) #k = isfinite(LB(i)) + 2*isfinite(UB(i));
        params['BoundClass'][i] = k #params.BoundClass(i) = k;
        if (k==3) and (LB[i]==UB[i]):#if (k==3) && (LB(i)==UB(i))
            params['BoundClass'][i] = 4 #params.BoundClass(i) = 4;
        #end
    #end

    #% transform starting values into their unconstrained
    #% surrogates. Check for infeasible starting guesses.
    x0u = x0 #x0u = x0;
    k=1 #k=1;
    for i in range(n): # i = 1:n
        match params['BoundClass'][i]: #switch params.BoundClass(i)
            case 1:
                #% lower bound only
                if x0[i]<=LB[i]: #x0(i)<=LB(i)
                    #% infeasible starting value. Use bound.
                    x0u[k] = 0
                else:
                    x0u[k] = np.sqrt(x0[i] - LB[i]) #x0u(k) = sqrt(x0(i) - LB(i));
                #end
                #% increment k
                k = k+1 #k=k+1;
            case 2:
                #% upper bound only
                if x0[i]>=UB[i]: #x0(i)>=UB(i):
                    #% infeasible starting value. use bound.
                    x0u[k] = 0 #x0u(k) = 0;
                else:
                    x0u[k] = np.sqrt(UB[i] - x0[i]) #x0u(k) = sqrt(UB(i) - x0(i));
                #end
      
                #% increment k
                k=k+1 #k=k+1;
            case 3:
                #% lower and upper bounds
                if x0[i] <= LB[i]: #x0(i)<=LB(i)
                    #% infeasible starting value
                    x0u[k] = -np.pi/2 #x0u(k) = -pi/2;
                elif x0[i] >= UB[i] : #x0(i)>=UB(i)
                    #% infeasible starting value
                    x0u[k] = np.pi/2 #x0u(k) = pi/2;
                else:
                    x0u[k] = 2*(x0[i] - LB[i])/(UB[i]-LB[i]) - 1#x0u(k) = 2*(x0(i) - LB(i))/(UB(i)-LB(i)) - 1;
                    #% shift by 2*pi to avoid problems at zero in fminsearch
                    #% otherwise, the initial simplex is vanishingly small
                    x0u[k] = 2*np.pi+np.asin(max(-1,min(1,x0u[k]))) #x0u(k) = 2*pi+asin(max(-1,min(1,x0u(k))));
                #end
      
                #% increment k
                k=k+1 #k=k+1;
            case 0:
                #% unconstrained variable. x0u(i) is set.
                x0u[k] = x0[i] #x0u(k) = x0(i);
      
                #% increment k
                k=k+1 #k=k+1;
            case 4:
                #% fixed variable. drop it before fminsearch sees it.
                #% k is not incremented for this variable.
                k = k
        #end
    #end
    #% if any of the unknowns were fixed, then we need to shorten
    #% x0u now.
    
    if (k <= n) : #k<=n
        x0u[k:n] = [] #x0u(k:n) = [];
    #end
        
    output = []
    #% were all the variables fixed?
    if len(x0u)==0: #isempty(x0u)
        #% All variables were fixed. quit immediately, setting the
        #% appropriate parameters, then return.
  
        #% undo the variable transformations into the original space
        x = xtransform(x0u,params) #x = xtransform(x0u,params);
  
        #% final reshape
        x = np.reshape(x,xsize)
  
        #% stuff fval with the final value
        fval = params['fun'](x,params['args']) #fval = feval(params.fun,x,params.args{:});
  
        #% fminsearchbnd was not called
        exitflag = 0
  
        output['iterations'] = 0 #output.iterations = 0;
        output['funcCount'] = 1 #output.funcCount = 1;
        output['algorithm'] = 'fminsearch' #output.algorithm = 'fminsearch';
        output['message'] = 'All variables were held fixed by the applied bounds' #output.message = 'All variables were held fixed by the applied bounds';
  
        #% return with no call at all to fminsearch
        return([x,fval,exitflag,output])
    #end

    #% Check for an outputfcn. If there is any, then substitute my
    #% own wrapper function.
    if len(options['OutputFcn']): #~isempty(options.OutputFcn)
        params['OutputFcn'] = options['OutputFcn'] #params.OutputFcn = options.OutputFcn;
        options['OutputFcn'] = outfun_wrapper #options.OutputFcn = @outfun_wrapper;
    #end

    ##????????????????????????????????????????????    
    ######################################    
    #% now we can call fminsearch, but with our own
    #% intra-objective function.
    [xu,fval,exitflag,output] = fminsearch(intrafun,x0u,options,params) #[xu,fval,exitflag,output] = fminsearch(@intrafun,x0u,options,params);

    #% undo the variable transformations into the original space
    x = xtransform(xu,params)

    #% final reshape to make sure the result has the proper shape
    x = np.reshape(x,xsize) #x = reshape(x,xsize);

    #% Use a nested function as the OutputFcn wrapper
    def outfun_wrapper(x,varargin): #function stop = outfun_wrapper(x,varargin);
        #% we need to transform x first
        xtrans = xtransform(x,params) #xtrans = xtransform(x,params);
    
        #% then call the user supplied OutputFcn
        stop = params['OutputFcn'](xtrans,varargin[1:(len(varargin)-1)]) #stop = params.OutputFcn(xtrans,varargin{1:(end-1)});
        return(stop)
    #end

#end % mainline end
    




    





# def get_PC_row_curr_ID(mst_ID):
#     curr_ID = 0

#     # Standard rows
#     if mst_ID == pcc.PC_PROG_SPEC_HR_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_PROG_SPEC_HR_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_NAT_LEVEL_STAFF_ROW_MST_ID:
#         curr_ID = pcc.PC_NAT_LEVEL_STAFF_ROW_CURR_ID

#     elif mst_ID == pcc.PC_REG_LEVEL_STAFF_ROW_MST_ID:
#         curr_ID = pcc.PC_REG_LEVEL_STAFF_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DIST_LEVEL_STAFF_ROW_MST_ID:
#         curr_ID = pcc.PC_DIST_LEVEL_STAFF_ROW_CURR_ID

#     elif mst_ID == pcc.PC_TRAIN_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_TRAIN_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_INSERV_REFRESH_TRAIN_ROW_MST_ID:
#         curr_ID = pcc.PC_INSERV_REFRESH_TRAIN_ROW_CURR_ID

#     elif mst_ID == pcc.PC_TRAIN_OF_TRAIN_ROW_MST_ID:
#         curr_ID = pcc.PC_TRAIN_OF_TRAIN_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DEV_TRAIN_PROG_MAT_ROW_MST_ID:
#         curr_ID = pcc.PC_DEV_TRAIN_PROG_MAT_ROW_CURR_ID

#     elif mst_ID == pcc.PC_CHANGE_PRESERV_TRAIN_CURRIC_ROW_MST_ID:
#         curr_ID = pcc.PC_CHANGE_PRESERV_TRAIN_CURRIC_ROW_CURR_ID

#     elif mst_ID == pcc.PC_SUPP_ACT_ROW_MST_ID:
#         curr_ID = pcc.PC_SUPP_ACT_ROW_CURR_ID

#     elif mst_ID == pcc.PC_SUPERVIS_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_SUPERVIS_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_COORD_MEET_ROW_MST_ID:
#         curr_ID = pcc.PC_COORD_MEET_ROW_CURR_ID

#     elif mst_ID == pcc.PC_NAT_STAFF_VISIT_LOCAL_ROW_MST_ID:
#         curr_ID = pcc.PC_NAT_STAFF_VISIT_LOCAL_ROW_CURR_ID

#     elif mst_ID == pcc.PC_MONITOR_EVAL_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_MONITOR_EVAL_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DESIGN_ME_FRAME_SYS_ROW_MST_ID:
#         curr_ID = pcc.PC_DESIGN_ME_FRAME_SYS_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DESIGN_QC_QA_ROW_MST_ID:
#         curr_ID = pcc.PC_DESIGN_QC_QA_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DESIGN_REV_DATA_MGMT_SYS_ROW_MST_ID:
#         curr_ID = pcc.PC_DESIGN_REV_DATA_MGMT_SYS_ROW_CURR_ID
 
#     elif mst_ID == pcc.PC_DATA_COLLECT_ANALYSIS_ROW_MST_ID:
#         curr_ID = pcc.PC_DATA_COLLECT_ANALYSIS_ROW_CURR_ID

#     elif mst_ID == pcc.PC_QC_QA_ROW_MST_ID:
#         curr_ID = pcc.PC_QC_QA_ROW_CURR_ID

#     elif mst_ID == pcc.PC_INFRA_EQUIP_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_INFRA_EQUIP_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_SIT_ASSESS_INFRA_ROW_MST_ID:
#         curr_ID = pcc.PC_SIT_ASSESS_INFRA_ROW_CURR_ID

#     elif mst_ID == pcc.PC_EQUIP_UPGRADES_LOW_TIER_ROW_MST_ID:
#         curr_ID = pcc.PC_EQUIP_UPGRADES_LOW_TIER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_EQUIP_UPGRADES_HOSP_ROW_MST_ID:
#         curr_ID = pcc.PC_EQUIP_UPGRADES_HOSP_ROW_CURR_ID

#     elif mst_ID == pcc.PC_TRANSPORT_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_TRANSPORT_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_SIT_ASSESS_TRANSPORT_ROW_MST_ID:
#         curr_ID = pcc.PC_SIT_ASSESS_TRANSPORT_ROW_CURR_ID

#     elif mst_ID == pcc.PC_NEW_VEHIC_PURCHASE_ROW_MST_ID:
#         curr_ID = pcc.PC_NEW_VEHIC_PURCHASE_ROW_CURR_ID

#     elif mst_ID == pcc.PC_VEHIC_OP_MAINT_ROW_MST_ID:
#         curr_ID = pcc.PC_VEHIC_OP_MAINT_ROW_CURR_ID

#     elif mst_ID == pcc.PC_COMM_MEDIA_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_COMM_MEDIA_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DEV_COMM_STRAT_ROW_MST_ID:
#         curr_ID = pcc.PC_DEV_COMM_STRAT_ROW_CURR_ID

#     elif mst_ID == pcc.PC_MASS_MEDIA_COMM_ROW_MST_ID:
#         curr_ID = pcc.PC_MASS_MEDIA_COMM_ROW_CURR_ID

#     elif mst_ID == pcc.PC_PRINTED_MATS_ROW_MST_ID:
#         curr_ID = pcc.PC_PRINTED_MATS_ROW_CURR_ID

#     elif mst_ID == pcc.PC_SOCIAL_OUTR_ACTIV_ROW_MST_ID:
#         curr_ID = pcc.PC_SOCIAL_OUTR_ACTIV_ROW_CURR_ID

#     elif mst_ID == pcc.PC_ADVOC_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_ADVOC_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_PLAN_ADVOC_STRAT_ROW_MST_ID:
#         curr_ID = pcc.PC_PLAN_ADVOC_STRAT_ROW_CURR_ID

#     elif mst_ID == pcc.PC_ADVOC_ACTS_ROW_MST_ID:
#         curr_ID = pcc.PC_ADVOC_ACTS_ROW_CURR_ID

#     elif mst_ID == pcc.PC_ADVOC_MATS_ROW_MST_ID:
#         curr_ID = pcc.PC_ADVOC_MATS_ROW_CURR_ID

#     elif mst_ID == pcc.PC_GEN_PROG_MGMT_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_GEN_PROG_MGMT_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DESIGN_REV_COUNTRY_STRAT_ROW_MST_ID:
#         curr_ID = pcc.PC_DESIGN_REV_COUNTRY_STRAT_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DEV_REVIEW_ANNUAL_WORKPLAN_ROW_MST_ID:
#         curr_ID = pcc.PC_DEV_REVIEW_ANNUAL_WORKPLAN_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DEV_REVIEW_HR_PLAN_ROW_MST_ID:
#         curr_ID = pcc.PC_DEV_REVIEW_HR_PLAN_ROW_CURR_ID

#     elif mst_ID == pcc.PC_PROG_COORD_MEET_ROW_MST_ID:
#         curr_ID = pcc.PC_PROG_COORD_MEET_ROW_CURR_ID

#     elif mst_ID == pcc.PC_COMMOD_REG_POL_ROW_MST_ID:
#         curr_ID = pcc.PC_COMMOD_REG_POL_ROW_CURR_ID

#     elif mst_ID == pcc.PC_SIT_ANALYSIS_ROW_MST_ID:
#         curr_ID = pcc.PC_SIT_ANALYSIS_ROW_CURR_ID

#     elif mst_ID == pcc.PC_OTHER_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_OFFICE_EQUIP_SUPP_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID:
#         curr_ID = pcc.PC_DBL_CLICK_ADD_OTHER_ITEMS_ROW_CURR_ID 

#     elif mst_ID == pcc.PC_OFFICE_EQUIP_SUPP_ROW_MST_ID:
#         curr_ID = pcc.PC_OTHER_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_UTIL_ROW_MST_ID:
#         curr_ID = pcc.PC_UTIL_ROW_CURR_ID

#     # HIV rows

#     elif mst_ID == pcc.PC_POLICY_BASED_INTERV_ROW_MST_ID:
#         curr_ID = pcc.PC_POLICY_BASED_INTERV_ROW_CURR_ID

#     elif mst_ID == pcc.PC_SOC_ECON_SUPP_CHILD_ROW_MST_ID:
#         curr_ID = pcc.PC_SOC_ECON_SUPP_CHILD_ROW_CURR_ID

#     elif mst_ID == pcc.PC_PREV_VIOL_WOMEN_ROW_MST_ID: 
#         curr_ID = pcc.PC_PREV_VIOL_WOMEN_ROW_CURR_ID

#     elif mst_ID == pcc.PC_DECRIM_KEY_POP_BEHAV_ROW_MST_ID:
#         curr_ID = pcc.PC_DECRIM_KEY_POP_BEHAV_ROW_CURR_ID

#     elif mst_ID == pcc.PC_PREV_STIGMA_DISCRIM_ROW_MST_ID:
#         curr_ID = pcc.PC_PREV_STIGMA_DISCRIM_ROW_CURR_ID

#     elif mst_ID == pcc.PC_COMM_CIVIL_SOC_LEAD_ROW_MST_ID:
#         curr_ID = pcc.PC_COMM_CIVIL_SOC_LEAD_ROW_CURR_ID

#     elif mst_ID == pcc.PC_LEGAL_POLICY_REFORMS_ROW_MST_ID:
#         curr_ID = pcc.PC_LEGAL_POLICY_REFORMS_ROW_CURR_ID

#     elif mst_ID == pcc.PC_HIV_OTHER_HEADER_ROW_MST_ID:
#         curr_ID = pcc.PC_HIV_OTHER_HEADER_ROW_CURR_ID

#     elif mst_ID == pcc.PC_HIV_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID:
#         curr_ID = pcc.PC_HIV_DBL_CLICK_ADD_OTHER_ITEMS_ROW_CURR_ID

#     return curr_ID

# # Need the programme area ID because some programme areas have different rows and the current IDs can overlap, unlike the master IDs.
# def get_PC_row_mst_ID(prog_area_ID = str, curr_ID = int):
#     mst_ID = 0

#     if (prog_area_ID == pccl.PC_IH_HIV_AIDS_MST_ID) and (curr_ID >= pcc.PC_POLICY_BASED_INTERV_ROW_CURR_ID):

#         if curr_ID == pcc.PC_POLICY_BASED_INTERV_ROW_CURR_ID:
#             mst_ID = pcc.PC_POLICY_BASED_INTERV_ROW_MST_ID

#         elif curr_ID == pcc.PC_SOC_ECON_SUPP_CHILD_ROW_CURR_ID:
#             mst_ID = pcc.PC_SOC_ECON_SUPP_CHILD_ROW_MST_ID

#         elif curr_ID == pcc.PC_PREV_VIOL_WOMEN_ROW_CURR_ID: 
#             mst_ID = pcc.PC_PREV_VIOL_WOMEN_ROW_MST_ID

#         elif curr_ID == pcc.PC_DECRIM_KEY_POP_BEHAV_ROW_CURR_ID:
#             mst_ID = pcc.PC_DECRIM_KEY_POP_BEHAV_ROW_MST_ID

#         elif curr_ID == pcc.PC_PREV_STIGMA_DISCRIM_ROW_CURR_ID:
#             mst_ID = pcc.PC_PREV_STIGMA_DISCRIM_ROW_MST_ID

#         elif curr_ID == pcc.PC_COMM_CIVIL_SOC_LEAD_ROW_CURR_ID:
#             mst_ID = pcc.PC_COMM_CIVIL_SOC_LEAD_ROW_MST_ID

#         elif curr_ID == pcc.PC_LEGAL_POLICY_REFORMS_ROW_CURR_ID:
#             mst_ID = pcc.PC_LEGAL_POLICY_REFORMS_ROW_MST_ID

#         elif curr_ID == pcc.PC_HIV_OTHER_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_HIV_OTHER_HEADER_ROW_MST_ID           

#         elif curr_ID == pcc.PC_HIV_DBL_CLICK_ADD_OTHER_ITEMS_ROW_CURR_ID:
#             mst_ID = pcc.PC_HIV_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID

#     elif (prog_area_ID != pccl.PC_IH_HIV_AIDS_MST_ID) or (curr_ID < pcc.PC_SOC_ECON_SUPP_CHILD_ROW_CURR_ID):

#         if curr_ID == pcc.PC_PROG_SPEC_HR_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_PROG_SPEC_HR_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_NAT_LEVEL_STAFF_ROW_CURR_ID:
#             mst_ID = pcc.PC_NAT_LEVEL_STAFF_ROW_MST_ID

#         elif curr_ID == pcc.PC_REG_LEVEL_STAFF_ROW_CURR_ID:
#             mst_ID = pcc.PC_REG_LEVEL_STAFF_ROW_MST_ID

#         elif curr_ID == pcc.PC_DIST_LEVEL_STAFF_ROW_CURR_ID:
#             mst_ID = pcc.PC_DIST_LEVEL_STAFF_ROW_MST_ID

#         elif curr_ID == pcc.PC_TRAIN_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_TRAIN_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_INSERV_REFRESH_TRAIN_ROW_CURR_ID:
#             mst_ID = pcc.PC_INSERV_REFRESH_TRAIN_ROW_MST_ID

#         elif curr_ID == pcc.PC_TRAIN_OF_TRAIN_ROW_CURR_ID:
#             mst_ID = pcc.PC_TRAIN_OF_TRAIN_ROW_MST_ID

#         elif curr_ID == pcc.PC_DEV_TRAIN_PROG_MAT_ROW_CURR_ID:
#             mst_ID = pcc.PC_DEV_TRAIN_PROG_MAT_ROW_MST_ID

#         elif curr_ID == pcc.PC_CHANGE_PRESERV_TRAIN_CURRIC_ROW_CURR_ID:
#             mst_ID = pcc.PC_CHANGE_PRESERV_TRAIN_CURRIC_ROW_MST_ID

#         elif curr_ID == pcc.PC_SUPP_ACT_ROW_CURR_ID:
#             mst_ID = pcc.PC_SUPP_ACT_ROW_MST_ID

#         elif curr_ID == pcc.PC_SUPERVIS_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_SUPERVIS_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_COORD_MEET_ROW_CURR_ID:
#             mst_ID = pcc.PC_COORD_MEET_ROW_MST_ID

#         elif curr_ID == pcc.PC_NAT_STAFF_VISIT_LOCAL_ROW_CURR_ID:
#             mst_ID = pcc.PC_NAT_STAFF_VISIT_LOCAL_ROW_MST_ID

#         elif curr_ID == pcc.PC_MONITOR_EVAL_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_MONITOR_EVAL_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_DESIGN_ME_FRAME_SYS_ROW_CURR_ID:
#             mst_ID = pcc.PC_DESIGN_ME_FRAME_SYS_ROW_MST_ID

#         elif curr_ID == pcc.PC_DESIGN_QC_QA_ROW_CURR_ID:
#             mst_ID = pcc.PC_DESIGN_QC_QA_ROW_MST_ID

#         elif curr_ID == pcc.PC_DESIGN_REV_DATA_MGMT_SYS_ROW_CURR_ID:
#             mst_ID = pcc.PC_DESIGN_REV_DATA_MGMT_SYS_ROW_MST_ID
    
#         elif curr_ID == pcc.PC_DATA_COLLECT_ANALYSIS_ROW_CURR_ID:
#             mst_ID = pcc.PC_DATA_COLLECT_ANALYSIS_ROW_MST_ID

#         elif curr_ID == pcc.PC_QC_QA_ROW_CURR_ID:
#             mst_ID = pcc.PC_QC_QA_ROW_MST_ID

#         elif curr_ID == pcc.PC_INFRA_EQUIP_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_INFRA_EQUIP_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_SIT_ASSESS_INFRA_ROW_CURR_ID:
#             mst_ID = pcc.PC_SIT_ASSESS_INFRA_ROW_MST_ID

#         elif curr_ID == pcc.PC_EQUIP_UPGRADES_LOW_TIER_ROW_CURR_ID:
#             mst_ID = pcc.PC_EQUIP_UPGRADES_LOW_TIER_ROW_MST_ID

#         elif curr_ID == pcc.PC_EQUIP_UPGRADES_HOSP_ROW_CURR_ID:
#             mst_ID = pcc.PC_EQUIP_UPGRADES_HOSP_ROW_MST_ID

#         elif curr_ID == pcc.PC_TRANSPORT_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_TRANSPORT_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_SIT_ASSESS_TRANSPORT_ROW_CURR_ID:
#             mst_ID = pcc.PC_SIT_ASSESS_TRANSPORT_ROW_MST_ID

#         elif curr_ID == pcc.PC_NEW_VEHIC_PURCHASE_ROW_CURR_ID:
#             mst_ID = pcc.PC_NEW_VEHIC_PURCHASE_ROW_MST_ID

#         elif curr_ID == pcc.PC_VEHIC_OP_MAINT_ROW_CURR_ID:
#             mst_ID = pcc.PC_VEHIC_OP_MAINT_ROW_MST_ID

#         elif curr_ID == pcc.PC_COMM_MEDIA_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_COMM_MEDIA_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_DEV_COMM_STRAT_ROW_CURR_ID:
#             mst_ID = pcc.PC_DEV_COMM_STRAT_ROW_MST_ID

#         elif curr_ID == pcc.PC_MASS_MEDIA_COMM_ROW_CURR_ID:
#             mst_ID = pcc.PC_MASS_MEDIA_COMM_ROW_MST_ID

#         elif curr_ID == pcc.PC_PRINTED_MATS_ROW_CURR_ID:
#             mst_ID = pcc.PC_PRINTED_MATS_ROW_MST_ID

#         elif curr_ID == pcc.PC_SOCIAL_OUTR_ACTIV_ROW_CURR_ID:
#             mst_ID = pcc.PC_SOCIAL_OUTR_ACTIV_ROW_MST_ID

#         elif curr_ID == pcc.PC_ADVOC_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_ADVOC_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_PLAN_ADVOC_STRAT_ROW_CURR_ID:
#             mst_ID = pcc.PC_PLAN_ADVOC_STRAT_ROW_MST_ID

#         elif curr_ID == pcc.PC_ADVOC_ACTS_ROW_CURR_ID:
#             mst_ID = pcc.PC_ADVOC_ACTS_ROW_MST_ID

#         elif curr_ID == pcc.PC_ADVOC_MATS_ROW_CURR_ID:
#             mst_ID = pcc.PC_ADVOC_MATS_ROW_MST_ID

#         elif curr_ID == pcc.PC_GEN_PROG_MGMT_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_GEN_PROG_MGMT_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_DESIGN_REV_COUNTRY_STRAT_ROW_CURR_ID:
#             mst_ID = pcc.PC_DESIGN_REV_COUNTRY_STRAT_ROW_MST_ID

#         elif curr_ID == pcc.PC_DEV_REVIEW_ANNUAL_WORKPLAN_ROW_CURR_ID:
#             mst_ID = pcc.PC_DEV_REVIEW_ANNUAL_WORKPLAN_ROW_MST_ID

#         elif curr_ID == pcc.PC_DEV_REVIEW_HR_PLAN_ROW_CURR_ID:
#             mst_ID = pcc.PC_DEV_REVIEW_HR_PLAN_ROW_MST_ID

#         elif curr_ID == pcc.PC_PROG_COORD_MEET_ROW_CURR_ID:
#             mst_ID = pcc.PC_PROG_COORD_MEET_ROW_MST_ID

#         elif curr_ID == pcc.PC_COMMOD_REG_POL_ROW_CURR_ID:
#             mst_ID = pcc.PC_COMMOD_REG_POL_ROW_MST_ID

#         elif curr_ID == pcc.PC_SIT_ANALYSIS_ROW_CURR_ID:
#             mst_ID = pcc.PC_SIT_ANALYSIS_ROW_MST_ID

#         elif curr_ID == pcc.PC_OTHER_HEADER_ROW_CURR_ID:
#             mst_ID = pcc.PC_OTHER_HEADER_ROW_MST_ID

#         elif curr_ID == pcc.PC_DBL_CLICK_ADD_OTHER_ITEMS_ROW_CURR_ID:
#             mst_ID = pcc.PC_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID 

#         elif curr_ID == pcc.PC_OFFICE_EQUIP_SUPP_ROW_CURR_ID:
#             mst_ID = pcc.PC_OFFICE_EQUIP_SUPP_ROW_MST_ID

#         elif curr_ID == pcc.PC_UTIL_ROW_CURR_ID:
#             mst_ID = pcc.PC_UTIL_ROW_MST_ID

#     return mst_ID

# def get_PC_row_name(mst_ID):
#     name = ''

#     # Standard rows

#     if mst_ID == pcc.PC_PROG_SPEC_HR_HEADER_ROW_MST_ID:
#         name = '1. Programme-specific human resources'

#     elif mst_ID == pcc.PC_NAT_LEVEL_STAFF_ROW_MST_ID:
#         name = '1.1 National-level staff'

#     elif mst_ID == pcc.PC_REG_LEVEL_STAFF_ROW_MST_ID:
#         name = '1.2 Regional-level staff'

#     elif mst_ID == pcc.PC_DIST_LEVEL_STAFF_ROW_MST_ID:
#         name = '1.3 District-level staff'

#     elif mst_ID == pcc.PC_TRAIN_HEADER_ROW_MST_ID:
#         name = '2. Training'

#     elif mst_ID == pcc.PC_INSERV_REFRESH_TRAIN_ROW_MST_ID:
#         name = '2.1 In-service / Refresher training'

#     elif mst_ID == pcc.PC_TRAIN_OF_TRAIN_ROW_MST_ID:
#         name = '2.2 Training of trainers'

#     elif mst_ID == pcc.PC_DEV_TRAIN_PROG_MAT_ROW_MST_ID:
#         name = '2.3 Development of training programmes and material'

#     elif mst_ID == pcc.PC_CHANGE_PRESERV_TRAIN_CURRIC_ROW_MST_ID:
#         name = '2.4 Changing the pre-service training curriculum'

#     elif mst_ID == pcc.PC_SUPP_ACT_ROW_MST_ID:
#         name = '2.5 Support activities'

#     elif mst_ID == pcc.PC_SUPERVIS_HEADER_ROW_MST_ID:
#         name = '3. Supervision'

#     elif mst_ID == pcc.PC_COORD_MEET_ROW_MST_ID:
#         name = '3.1 Coordination meetings'

#     elif mst_ID == pcc.PC_NAT_STAFF_VISIT_LOCAL_ROW_MST_ID:
#         name = '3.2 National staff visiting local staff'

#     elif mst_ID == pcc.PC_MONITOR_EVAL_HEADER_ROW_MST_ID:
#         name = '4. Monitoring and evaluation'

#     elif mst_ID == pcc.PC_DESIGN_ME_FRAME_SYS_ROW_MST_ID:
#         name = '4.1 Design of M & E frameworks and systems'

#     elif mst_ID == pcc.PC_DESIGN_QC_QA_ROW_MST_ID:
#         name = '4.2 Design of quality control and assurance'

#     elif mst_ID == pcc.PC_DESIGN_REV_DATA_MGMT_SYS_ROW_MST_ID:
#         name = '4.3 Design / review of data management systems'
 
#     elif mst_ID == pcc.PC_DATA_COLLECT_ANALYSIS_ROW_MST_ID:
#         name = '4.4 Data collection and analysis'

#     elif mst_ID == pcc.PC_QC_QA_ROW_MST_ID:
#         name = '4.5 Quality control / quality assurance'

#     elif mst_ID == pcc.PC_INFRA_EQUIP_HEADER_ROW_MST_ID:
#         name = '5. Infrastructure and equipment'

#     elif mst_ID == pcc.PC_SIT_ASSESS_INFRA_ROW_MST_ID:
#         name = '5.1  Situational assessment'

#     elif mst_ID == pcc.PC_EQUIP_UPGRADES_LOW_TIER_ROW_MST_ID:
#         name = '5.2 Equipment upgrades for lower-tier facilities'

#     elif mst_ID == pcc.PC_EQUIP_UPGRADES_HOSP_ROW_MST_ID:
#         name = '5.3 Equipment upgrades for hospitals'

#     elif mst_ID == pcc.PC_TRANSPORT_HEADER_ROW_MST_ID:
#         name = '6. Transport'

#     elif mst_ID == pcc.PC_SIT_ASSESS_TRANSPORT_ROW_MST_ID:
#         name = '6.1 Situational assessment'

#     elif mst_ID == pcc.PC_NEW_VEHIC_PURCHASE_ROW_MST_ID:
#         name = '6.2 New vehicle purchase'

#     elif mst_ID == pcc.PC_VEHIC_OP_MAINT_ROW_MST_ID:
#         name = '6.3 Vehicle operation and maintenance'

#     elif mst_ID == pcc.PC_COMM_MEDIA_HEADER_ROW_MST_ID:
#         name = '7. Communication, media, and outreach'

#     elif mst_ID == pcc.PC_DEV_COMM_STRAT_ROW_MST_ID:
#         name = '7.1 Development of communication strategy'

#     elif mst_ID == pcc.PC_MASS_MEDIA_COMM_ROW_MST_ID:
#         name = '7.2 Mass media'

#     elif mst_ID == pcc.PC_PRINTED_MATS_ROW_MST_ID:
#         name = '7.3 Printed materials'

#     elif mst_ID == pcc.PC_SOCIAL_OUTR_ACTIV_ROW_MST_ID:
#         name = '7.4 Social outreach activities'

#     elif mst_ID == pcc.PC_ADVOC_HEADER_ROW_MST_ID:
#         name = '8. Advocacy'

#     elif mst_ID == pcc.PC_PLAN_ADVOC_STRAT_ROW_MST_ID:
#         name = '8.1 Planning on advocacy strategy'

#     elif mst_ID == pcc.PC_ADVOC_ACTS_ROW_MST_ID:
#         name = '8.2 Advocacy activities'

#     elif mst_ID == pcc.PC_ADVOC_MATS_ROW_MST_ID:
#         name = '8.3 Advocacy materials'

#     elif mst_ID == pcc.PC_GEN_PROG_MGMT_HEADER_ROW_MST_ID:
#         name = '9. General programme management'

#     elif mst_ID == pcc.PC_DESIGN_REV_COUNTRY_STRAT_ROW_MST_ID:
#         name = '9.1 Design and review of country strategy'

#     elif mst_ID == pcc.PC_DEV_REVIEW_ANNUAL_WORKPLAN_ROW_MST_ID:
#         name = '9.2 Development and review of annual work plan'

#     elif mst_ID == pcc.PC_DEV_REVIEW_HR_PLAN_ROW_MST_ID:
#         name = '9.3 Development and review of human resources plan'

#     elif mst_ID == pcc.PC_PROG_COORD_MEET_ROW_MST_ID:
#         name = '9.4 Programme coordination meetings'

#     elif mst_ID == pcc.PC_COMMOD_REG_POL_ROW_MST_ID:
#         name = '9.5 Commodity regulation and policies'

#     elif mst_ID == pcc.PC_SIT_ANALYSIS_ROW_MST_ID:
#         name = '9.6 Situation analysis'

#     elif mst_ID == pcc.PC_OFFICE_EQUIP_SUPP_ROW_MST_ID:
#         name = '9.7 Office equipment and supplies'
        
#     elif mst_ID == pcc.PC_UTIL_ROW_MST_ID:
#         name = '9.8 Utilities'

#     elif mst_ID == pcc.PC_OTHER_HEADER_ROW_MST_ID:
#         name = '10. Other'

#     elif mst_ID == pcc.PC_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID:
#         name = '10.1 Other items'

#     # HIV-specific rows

#     elif mst_ID == pcc.PC_POLICY_BASED_INTERV_ROW_MST_ID:
#         name = '10. Policy-based interventions'      

#     elif mst_ID == pcc.PC_SOC_ECON_SUPP_CHILD_ROW_MST_ID:
#         name = '10.1 Socioeconomic support for orphans and vulnerable children'         

#     elif mst_ID == pcc.PC_PREV_VIOL_WOMEN_ROW_MST_ID:           
#         name = '10.2 Prevention of violence against women'    

#     elif mst_ID == pcc.PC_DECRIM_KEY_POP_BEHAV_ROW_MST_ID:      
#         name = '10.3 Decriminalization of key population behaviors'   

#     elif mst_ID == pcc.PC_PREV_STIGMA_DISCRIM_ROW_MST_ID:       
#         name = '10.4 Prevention of stigma and discrimination'    

#     elif mst_ID == pcc.PC_COMM_CIVIL_SOC_LEAD_ROW_MST_ID:       
#         name = '10.5 Community and civil society leadership, including young people'    

#     elif mst_ID == pcc.PC_LEGAL_POLICY_REFORMS_ROW_MST_ID:       
#         name = '10.6 Legal, policy reforms'    

#     elif mst_ID == pcc.PC_HIV_OTHER_HEADER_ROW_MST_ID:                 
#         name = '11. Other'  

#     elif mst_ID == pcc.PC_HIV_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID:    
#         name = '11.1 Other items'

#     return name

# def get_PC_row_string_constant(mst_ID):
#     string_constant = ''

#     # Standard rows
#     if mst_ID == pcc.PC_PROG_SPEC_HR_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stprogSpecHR'

#     elif mst_ID == pcc.PC_NAT_LEVEL_STAFF_ROW_MST_ID:
#         string_constant = 'GB_stNatStaff'

#     elif mst_ID == pcc.PC_REG_LEVEL_STAFF_ROW_MST_ID:
#         string_constant = 'GB_stISProvinStaff'

#     elif mst_ID == pcc.PC_DIST_LEVEL_STAFF_ROW_MST_ID:
#         string_constant = 'GB_stDistStaff'

#     elif mst_ID == pcc.PC_TRAIN_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stTraining'

#     elif mst_ID == pcc.PC_INSERV_REFRESH_TRAIN_ROW_MST_ID:
#         string_constant = 'GB_stInServRefTrain'

#     elif mst_ID == pcc.PC_TRAIN_OF_TRAIN_ROW_MST_ID:
#         string_constant = 'GB_stTrainOfTrainers'

#     elif mst_ID == pcc.PC_DEV_TRAIN_PROG_MAT_ROW_MST_ID:
#         string_constant = 'GB_stDevTrainProgAndMat'

#     elif mst_ID == pcc.PC_CHANGE_PRESERV_TRAIN_CURRIC_ROW_MST_ID:
#         string_constant = 'GB_stChangePreServTrain'

#     elif mst_ID == pcc.PC_SUPP_ACT_ROW_MST_ID:
#         string_constant = 'GB_stSuppActivities'

#     elif mst_ID == pcc.PC_SUPERVIS_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stSupervision'

#     elif mst_ID == pcc.PC_COORD_MEET_ROW_MST_ID:
#         string_constant = 'GB_stCoordMeets'

#     elif mst_ID == pcc.PC_NAT_STAFF_VISIT_LOCAL_ROW_MST_ID:
#         string_constant = 'GB_stNatStaffVisLocal'

#     elif mst_ID == pcc.PC_MONITOR_EVAL_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stMonitAndEval'

#     elif mst_ID == pcc.PC_DESIGN_ME_FRAME_SYS_ROW_MST_ID:
#         string_constant = 'GB_stDesMEFramewrkSys'

#     elif mst_ID == pcc.PC_DESIGN_QC_QA_ROW_MST_ID:
#         string_constant = 'GB_stDesQCAssur'

#     elif mst_ID == pcc.PC_DESIGN_REV_DATA_MGMT_SYS_ROW_MST_ID:
#         string_constant = 'GB_stDesRevDataMgmt'
 
#     elif mst_ID == pcc.PC_DATA_COLLECT_ANALYSIS_ROW_MST_ID:
#         string_constant = 'GB_stDataCollectAna'

#     elif mst_ID == pcc.PC_QC_QA_ROW_MST_ID:
#         string_constant = 'GB_stMonitQualityControl'

#     elif mst_ID == pcc.PC_INFRA_EQUIP_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stInfraEquip'

#     elif mst_ID == pcc.PC_SIT_ASSESS_INFRA_ROW_MST_ID:
#         string_constant = 'GB_stInfraSituationalAsses'

#     elif mst_ID == pcc.PC_EQUIP_UPGRADES_LOW_TIER_ROW_MST_ID:
#         string_constant = 'GB_stEquipUpgradesLowTierFac'

#     elif mst_ID == pcc.PC_EQUIP_UPGRADES_HOSP_ROW_MST_ID:
#         string_constant = 'GB_stEquipUpgradesHospitals'

#     elif mst_ID == pcc.PC_TRANSPORT_HEADER_ROW_MST_ID:
#         string_constant = 'SM_stTransport'

#     elif mst_ID == pcc.PC_SIT_ASSESS_TRANSPORT_ROW_MST_ID:
#         string_constant = 'GB_stTransSituationalAsses'

#     elif mst_ID == pcc.PC_NEW_VEHIC_PURCHASE_ROW_MST_ID:
#         string_constant = 'GB_stNewVehicPurchase'

#     elif mst_ID == pcc.PC_VEHIC_OP_MAINT_ROW_MST_ID:
#         string_constant = 'GB_stVehicOpMaint'

#     elif mst_ID == pcc.PC_COMM_MEDIA_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stCommMediaOutrch'

#     elif mst_ID == pcc.PC_DEV_COMM_STRAT_ROW_MST_ID:
#         string_constant = 'GB_stDevCommStrat'

#     elif mst_ID == pcc.PC_MASS_MEDIA_COMM_ROW_MST_ID:
#         string_constant = 'GB_stMassMedia'

#     elif mst_ID == pcc.PC_PRINTED_MATS_ROW_MST_ID:
#         string_constant = 'GB_stPrintedMats'

#     elif mst_ID == pcc.PC_SOCIAL_OUTR_ACTIV_ROW_MST_ID:
#         string_constant = 'GB_stSocialOutreachActiv'

#     elif mst_ID == pcc.PC_ADVOC_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stAdvocacy'

#     elif mst_ID == pcc.PC_PLAN_ADVOC_STRAT_ROW_MST_ID:
#         string_constant = 'GB_stPannAdvocacyStrat'

#     elif mst_ID == pcc.PC_ADVOC_ACTS_ROW_MST_ID:
#         string_constant = 'GB_stAdvocacyActiv'

#     elif mst_ID == pcc.PC_ADVOC_MATS_ROW_MST_ID:
#         string_constant = 'GB_stAdvocacyMats'

#     elif mst_ID == pcc.PC_GEN_PROG_MGMT_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stGenProgMgmt'

#     elif mst_ID == pcc.PC_DESIGN_REV_COUNTRY_STRAT_ROW_MST_ID:
#         string_constant = 'GB_stDesRevCntyStrat'

#     elif mst_ID == pcc.PC_DEV_REVIEW_ANNUAL_WORKPLAN_ROW_MST_ID:
#         string_constant = 'GB_stDevRevAnnWrkPlan'

#     elif mst_ID == pcc.PC_DEV_REVIEW_HR_PLAN_ROW_MST_ID:
#         string_constant = 'GB_stDevRevHRPlan'

#     elif mst_ID == pcc.PC_PROG_COORD_MEET_ROW_MST_ID:
#         string_constant = 'GB_stProgCoordMeet'

#     elif mst_ID == pcc.PC_COMMOD_REG_POL_ROW_MST_ID:
#         string_constant = 'GB_stCommodityRegPolicyNoNum'

#     elif mst_ID == pcc.PC_SIT_ANALYSIS_ROW_MST_ID:
#         string_constant = 'GB_stPCSitAnaCoordGrp'

#     elif mst_ID == pcc.PC_OTHER_HEADER_ROW_MST_ID:
#         string_constant = 'GB_stOther'

#     elif mst_ID == pcc.PC_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID:
#         string_constant = 'GB_stAddOtherItems'

#     elif mst_ID == pcc.PC_OFFICE_EQUIP_SUPP_ROW_MST_ID:
#         string_constant = 'GB_stOfficeEquipSupplies'
        
#     elif mst_ID == pcc.PC_UTIL_ROW_MST_ID:
#         string_constant = 'GB_stUtilities'

#     # HIV-specific rows

#     elif mst_ID == pcc.PC_POLICY_BASED_INTERV_ROW_MST_ID:
#         string_constant = 'GB_stPolicyBasedInterventions'

#     elif mst_ID == pcc.PC_SOC_ECON_SUPP_CHILD_ROW_MST_ID:
#         string_constant = 'GB_stSocioeconSuppChildren'

#     elif mst_ID == pcc.PC_PREV_VIOL_WOMEN_ROW_MST_ID:           
#         string_constant = 'GB_stPrevViolenceAgainstWomen'       

#     elif mst_ID == pcc.PC_DECRIM_KEY_POP_BEHAV_ROW_MST_ID:      
#         string_constant = 'GB_stDecrimKeyPopBehav'           

#     elif mst_ID == pcc.PC_PREV_STIGMA_DISCRIM_ROW_MST_ID:       
#         string_constant = 'GB_stPrevStigmaDiscrim'    

#     elif mst_ID == pcc.PC_COMM_CIVIL_SOC_LEAD_ROW_MST_ID:       
#         string_constant = 'GB_stCommCivilSocLead'    

#     elif mst_ID == pcc.PC_LEGAL_POLICY_REFORMS_ROW_MST_ID:       
#         string_constant = 'GB_stLegalPolicyReforms'    

#     elif mst_ID == pcc.PC_HIV_OTHER_HEADER_ROW_MST_ID:                 
#         string_constant = 'GB_stOther'  

#     elif mst_ID == pcc.PC_HIV_DBL_CLICK_ADD_OTHER_ITEMS_ROW_MST_ID:    
#         string_constant = 'GB_stAddOtherItems'

#     return string_constant

# def get_PC_row_is_aggregate(mst_ID):
#     # HIV programme area has different rows at the end, but the master IDs don't overlap with the standard programme areas, so we can utilize
#     # the same function.
#     if (mst_ID == pcc.PC_PROG_SPEC_HR_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_TRAIN_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_SUPERVIS_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_MONITOR_EVAL_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_INFRA_EQUIP_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_TRANSPORT_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_COMM_MEDIA_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_ADVOC_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_GEN_PROG_MGMT_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_OTHER_HEADER_ROW_MST_ID) \
#         or (mst_ID == pcc.PC_HIV_OTHER_HEADER_ROW_MST_ID):
#             return True
#     else:
#         return False

# def get_PC_row_has_data(prog_area_ID = str, mst_ID = str):
#     if (prog_area_ID == pccl.PC_IH_HIV_AIDS_MST_ID) and (mst_ID == pcc.PC_POLICY_BASED_INTERV_ROW_MST_ID):
#         return False
#     else:
#         return True

# def get_PC_row_is_cost(prog_area_ID = str, mst_ID = str):
#     if (prog_area_ID == pccl.PC_IH_HIV_AIDS_MST_ID) and (mst_ID == pcc.PC_POLICY_BASED_INTERV_ROW_MST_ID):
#         return False
#     else:
#         return True

# def get_PC_row_is_single_line(prog_area_ID = str, mst_ID = str):
#     if ((prog_area_ID == pccl.PC_IH_HIV_AIDS_MST_ID) and ((mst_ID == pcc.PC_SOC_ECON_SUPP_CHILD_ROW_MST_ID) 
#         or (mst_ID == pcc.PC_PREV_VIOL_WOMEN_ROW_MST_ID) 
#         or (mst_ID == pcc.PC_DECRIM_KEY_POP_BEHAV_ROW_MST_ID) 
#         or (mst_ID == pcc.PC_PREV_STIGMA_DISCRIM_ROW_MST_ID) 
#         or (mst_ID == pcc.PC_PREV_STIGMA_DISCRIM_ROW_MST_ID) 
#         or (mst_ID == pcc.PC_COMM_CIVIL_SOC_LEAD_ROW_MST_ID) 
#         or (mst_ID == pcc.PC_LEGAL_POLICY_REFORMS_ROW_MST_ID))):
#             return True
#     else:
#         return False

# def get_PPLI_intensity_value(phase_ID):
#     intensity = 0

#     if phase_ID == pcc.PC_EARLY_IMP_PHASE_25_PPLI_MST_ID:
#         intensity = 25

#     elif phase_ID == pcc.PC_PARTIAL_IMP_PHASE_50_PPLI_MST_ID:
#         intensity = 50

#     elif phase_ID == pcc.PC_FULL_IMP_PHASE_100_PPLI_MST_ID:
#         intensity = 100

#     return intensity

# def get_PPLI_phase_curr_ID(phase_ID):
#     phase_curr_ID = gbc.GB_NOT_FOUND

#     if phase_ID == pcc.PC_NO_ACTION_PHASE_0_PPLI_MST_ID:
#         phase_curr_ID = pcc.PC_NO_ACTION_PHASE_0_PPLI_CURR_ID

#     elif phase_ID == pcc.PC_PLANNING_PHASE_0_PPLI_MST_ID:
#         phase_curr_ID = pcc.PC_PLANNING_PHASE_0_PPLI_CURR_ID

#     elif phase_ID == pcc.PC_EARLY_IMP_PHASE_25_PPLI_MST_ID:
#         phase_curr_ID = pcc.PC_EARLY_IMP_PHASE_25_PPLI_CURR_ID

#     elif phase_ID == pcc.PC_PARTIAL_IMP_PHASE_50_PPLI_MST_ID:
#         phase_curr_ID = pcc.PC_PARTIAL_IMP_PHASE_50_PPLI_CURR_ID

#     elif phase_ID == pcc.PC_FULL_IMP_PHASE_100_PPLI_MST_ID:
#         phase_curr_ID = pcc.PC_FULL_IMP_PHASE_100_PPLI_CURR_ID

#     return phase_curr_ID

# def PC_get_health_sys_admin_unit_ID(admin_level_curr_ID):
#     ID = str(gbc.GB_NOT_FOUND)

#     if admin_level_curr_ID == pcc.PC_PPLI_NATIONAL_ADMIN_LEVEL_CURR_ID:
#         ID = str(pccl.PC_IH_NATIONAL_HSAU_MST_ID)

#     elif admin_level_curr_ID == pcc.PC_PPLI_PROVINCE_ADMIN_LEVEL_CURR_ID:
#         ID = str(pccl.PC_IH_PROVINCE_HSAU_MST_ID)

#     elif admin_level_curr_ID == pcc.PC_PPLI_DISTRICT_ADMIN_LEVEL_CURR_ID:
#         ID = str(pccl.PC_IH_DISTRICT_HSAU_MST_ID)

#     return ID





################ GB ################

################ AM ################
AMDatabaseVersion = 'V8'
AMEasyAIMDatabaseVersion = 'V5'
AMCSAVRDatabaseVersion = 'V3'
AMPreventionsNeedsVersion = 'V4'

#  UNAIDS output file template 
AMUNAIDSSummaryTemplateVersion = 'V2'

DPDatabaseVersion = 'v0'
DPUPDVersion = 'V5'
DPInitialConditionsDBVersion = 'V9'

DPUADatabaseVersion = 'v0'


################ TB ################
TB_WHO_DB_VERSION = 'V6'
TB_FORT_INPUT_VERSION = 'V7'
TB_FORT_OUTPUT_VERSION = 'V7'
TB_DIAGNOSTICS_VERSION = 'V6'
TB_DYNAMICAL_MODEL_VERSION= 'V2'




import numpy as np
from scipy import sparse


def make_model3(p, r, i, s, gps,prm) : #function M = make_model(p, r, i, s, gps,prm)  % Kenya

    r = r.copy()
    r['progression'] = np.array(r['progression'])
    r['reactivation'] = np.array(r['reactivation'])
    r['r_sym'] = np.array(r['sym'])

        #% --- Get the linear rates ------------------------------------------------
    m  = np.zeros((i['nstates'],i['nstates'])) #m  = np.zeros(i.nstates);
    m2 = np.zeros((i['nstates'],i['nstates'])) #m2 = np.zeros(i.nstates);
    m4 = np.zeros((i['nstates'],i['nstates']))

    for ia in range(len(gps['age'])): #ia = 1:length(gps.age)
        age  = gps['age'][ia] #age  = gps.age{ia};

        for istr in range(len(gps['strains'])): #istr = 1:length(gps.strains)
            strain = gps['strains'][istr] #strain = gps.strains{istr};
    
            for ih in range(len(gps['hiv'])): #ih = 1:length(gps.hiv)
                hiv = gps['hiv'][ih] #hiv = gps.hiv{ih};
        
                #getst = @(st) i.(st).(age).(hiv).(strain) #getst = @(st) i.(st).(age).(hiv).(strain);
                Isc  = i['Isc'][age][hiv][strain] #Isc  = getst('Isc');     % Subclinical 
                I    = i['I'][age][hiv][strain] #I    = getst('I');       % clinical symptom 
                E    = i['E'][age][hiv][strain] #E    = getst('E');
                Dxs  = i['Dx'][age][hiv][strain] #Dxs  = getst('Dx');
                Txs  = i['Tx'][age][hiv][strain] #Txs  = getst('Tx');
                Tx2s = i['Tx2'][age][hiv][strain] #Tx2s = getst('Tx2');
                Rlo  = i['Rlo'][age][hiv][strain] #Rlo  = getst('Rlo');
                Rhi  = i['Rhi'][age][hiv][strain] #Rhi  = getst('Rhi');
                R    = i['R'][age][hiv][strain] #R    = getst('R');
      
                #% Group Selectors
                ismdr   = strain == 'mdr' #ismdr   = strcmp(strain, 'mdr');
        
                for iv in range(len(gps['vacc'])): #iv = 1:length(gps.vacc)
                    vacc = gps['vacc'][iv] #vacc = gps.vacc{iv};
                    #getst2 = @(st) i.(st).(age).(hiv).(vacc).(strain) #getst2 = @(st) i.(st).(age).(hiv).(vacc).(strain);
                    Lf = i['Lf'][age][hiv][vacc][strain]  #Lf = getst2('Lf') ;
                    Ls = i['Ls'][age][hiv][vacc][strain] #Ls = getst2('Ls');
            
                    #% --- Fast progression and LTBI stabilisation
                    source  = Lf #source  = Lf;
                    destins = np.concatenate([Isc,                                                   Ls]) #destins = [Isc,                                                   Ls];
                    rates   = [r['progression'][ia,ih]*p['PT_PLHIV'][ih]*(1-(iv==1)*p['VE'][1]), r['LTBI_stabil'][ih]] #rates   = [r.progression(ia,ih).*p.PT_PLHIV(ih)*(1-(iv==2)*p.VE(2)), r.LTBI_stabil(ih)];     % p.PT_PLHIV = [1 1 (1-0.6)] during PT among PLHIV
                    m4[destins, source] = m4[destins, source] + rates #m(destins, source) = m(destins, source) + rates';
            
                    #% --- Reactivation
                    source = Ls; destin = Isc; rate = r['reactivation'][ia,ih]*p['PT_PLHIV'][ih]*(1-(iv==1)*p['VE'][1]) #source = Ls; destin = Isc; rate = r.reactivation(ia,ih).*p.PT_PLHIV(ih)*(1-(iv==2)*p.VE(2));  % p.PT_PLHIV = [1 1 (1-0.6)] during PT among PLHIV
                    m4[destins, source] = m4[destins, source] + rate #m(destin, source) = m(destin, source) + rate;
           
                #end
        
                #% --- Sub-clinical to clinical Infection
                source = Isc #source = Isc;
                destin = I #               destin = I;
                rate   = r['r_sym']#rate   = r.sym;                                 % rate of symptom development (estimate this value)
                m[destin, source] = m[destin, source] + rate #m(destin, source) = m(destin, source) + rate;
        
                #% Introduce public/private sector
                #% --- Primary careseeking, including access to public sector care
                source = I; destin = Dxs['pu']; rate = r['cs'][ia]*p['pu'] + r['access'] #source = I; destin = Dxs.pu; rate = r.cs(ia)*p.pu + r.access;
                m2[destin, source] = m2[destin, source] + rate #m2(destin, source) = m2(destin, source) + rate;
        
                source = I; destin = Dxs['pr']; rate = r['cs'][ia]*(1-p['pu']) #source = I; destin = Dxs.pr; rate = r.cs(ia)*(1-p.pu);
                m2[destin, source] = m2[destin, source] + rate #m2(destin, source) = m2(destin, source) + rate;
        
                #% --- Secondary careseeking
                source = E; destin = Dxs['pu']; rate = r['cs2'][ia]*p['pu'] #source = E; destin = Dxs.pu; rate = r.cs2(ia)*p.pu;
                m2[destin, source] = m2[destin, source] + rate #m2(destin, source) = m2(destin, source) + rate;
        
                source = E; destin = Dxs['pr']; rate = r['cs2'][ia]*(1-p['pu']) #source = E; destin = Dxs.pr; rate = r.cs2(ia)*(1-p.pu);
                m2[destin, source] = m2[destin, source] + rate #m2(destin, source) = m2(destin, source) + rate;
        
                #% --- Case finding from subclinical TB
                source = Isc; destin = Dxs['pu']; rate = r['cs3'][ia]*p['pu'] #source = Isc; destin = Dxs.pu; rate = r.cs3(ia)*p.pu ;
                m2[destin, source] = m2[destin, source] + rate #m2(destin, source) = m2(destin, source) + rate;
        
                source = Isc; destin = Dxs['pr']; rate = r['cs3'][ia]*(1-p['pu']) #source = Isc; destin = Dxs.pr; rate = r.cs3(ia)*(1-p.pu);
                m2[destin, source] = m2[destin, source] + rate #m2(destin, source) = m2(destin, source) + rate;
  
        
                for ip in range(len(gps['provs'])): #ip = 1:length(gps.provs)
                    prov = gps['provs'][ip] #prov = gps.provs{ip};
            
                    Dx = Dxs[prov]; Tx = Txs[prov]; Tx2 = Tx2s[prov] #Dx = Dxs.(prov); Tx = Txs.(prov); Tx2 = Tx2s.(prov);
            
                    try:
                        pFLinit = p['Dx'][ip]*p['Tx_init'][ip]*(1 - ismdr*p['MDR_rec'][ip])#pFLinit = p.Dx(ip)*p.Tx_init(ip)*(1 - ismdr*p.MDR_rec(ip)); %*p.Tx_init2(ip)
                    except:
                        KeyboardInterrupt
                    #end
                    pSLinit = p['Dx'][ip]*p['Tx_init2'][ip]*ismdr*p['MDR_rec'][ip] #pSLinit = p.Dx(ip)*p.Tx_init2(ip)*ismdr*p.MDR_rec(ip);
                    p_ltfu  = 1-(pFLinit+pSLinit) #; %p.Dx(ip)*p.Tx_init(ip) #p_ltfu  = 1-(pFLinit+pSLinit); %p.Dx(ip)*p.Tx_init(ip);
            
                    source  = Dx #source  = Dx;
                    destins = np.concatenate([Tx,       Tx2,      E]) #destins =          [Tx,       Tx2,      E];
                    rates   = r['Dx']*np.array([pFLinit,  pSLinit,  p_ltfu]) #rates   = r.Dx*[pFLinit,  pSLinit,  p_ltfu];
                    if(np.shape(rates)==(3,1)):
                        rates = rates.T[0]
                    m[destins, source] = m[destins, source] + rates #m(destins, source) = m(destins, source) + rates';
            
                    #% --- FL Treatment
                    pFLcure = p['cure'][ip]*(1-ismdr) #pFLcure = p.cure(ip)*(1-ismdr);
                    pSLtran = p['SL_trans'][ip]*ismdr #pSLtran = p.SL_trans(ip)*ismdr;
                    rMDRacq = r['MDR_acqu']*(1-ismdr)#                    rMDRacq = r.MDR_acqu*(1-ismdr);
                    if(np.shape(rMDRacq)==(1,)):
                        rMDRacq = rMDRacq[0]
            
                    source  = Tx #source  = Tx;
                    destins = np.concatenate([Rlo,            Tx2,                        E,                              Rhi,             i['Tx'][age][hiv]['mdr'][prov]]) #destins = [Rlo            Tx2                        E                              Rhi,             i.Tx.(age).(hiv).mdr.(prov)];
                    rates   = [r['Tx']*pFLcure,  r['Tx']*(1-pFLcure)*pSLtran,  r['Tx']*(1-pFLcure)*(1-pSLtran),  r['default'][ip],   rMDRacq] #rates   = [r.Tx*pFLcure,  r.Tx*(1-pFLcure)*pSLtran,  r.Tx*(1-pFLcure)*(1-pSLtran),  r.default(ip),   rMDRacq];
                    m[destins, source] = m[destins, source] + rates #m(destins, source) = m(destins, source) + rates';
            
                    #% --- SL Treatment
                    source  = Tx2 #source  = Tx2;
                    destins = np.concatenate([Rlo,                 E]) #destins = [Rlo                 E];          %Rhi];
                    rates   = [r['Tx2']*p['cure2'][ip],  r['Tx2']*(1-p['cure2'][ip])+ r['default2'][ip]] #rates   = [r.Tx2*p.cure2(ip),  r.Tx2*(1-p.cure2(ip))+ r.default2(ip)];
                    m[destins, source] = m[destins, source] + rates#m(destins, source) = m(destins, source) + rates';
            
            
                #end  % provider loop
        
                #% --- Relapse
                sources = np.concatenate([Rlo, Rhi, R]) #sources = [Rlo, Rhi, R];
                destin  = Isc #destin  = Isc;
                rates   = r['relapse'] #rates   = r.relapse;
                m[destin, sources] = m[destin, sources] + rates
        
                sources = [Rlo, Rhi] #sources = [Rlo, Rhi];
                destin  = R#destin  = R;
                rates   = 0.5
                m[destin, sources] = m[destin, sources] + rates #m(destin, sources) = m(destin, sources) + rates;
        
                #% --- Self cure
                sources = np.intersect1d(np.intersect1d(np.intersect1d(s['infectious'],s[hiv]), s[strain]),s[age]) #sources = intersect(intersect(intersect(s.infectious,s.(hiv)), s.(strain)),s.(age)); 
                destin  = Rhi #destin  = Rhi;
                rates   = r['self_cure'][ih] #rates   = r.self_cure(ih);
                m[destin, sources] = m[destin, sources] + rates #m(destin, sources) = m(destin, sources) + rates;
        
            #end  % hiv loop
        #end % MDR loop

        for ih in range(len(gps['hiv'])) :#ih = 1:length(gps.hiv)
            hiv = gps['hiv'][ih] #hiv = gps.hiv{ih};
    
            source = i['U'][age][hiv]['v0']#source = i.U.(age).(hiv).v0;
            destin = i['U'][age][hiv]['v1'] #destin = i.U.(age).(hiv).v1;
            rate   = r['vacc'][ia] #rate   = r.vacc(ia);
            m[destin,source] = m[destin,source] + rate #m(destin,source) = m(destin,source) + rate;
    
            source = i['U'][age][hiv]['v1'] #source = i.U.(age).(hiv).v1;
            destin = i['U'][age][hiv]['v2'] #destin = i.U.(age).(hiv).v2;
            rate   = r['waning'] #rate   = r.waning;
            m[destin,source] = m[destin,source] + rate#m(destin,source) = m(destin,source) + rate;
    
            for istr in range(len(gps['strains'])): #istr = 1:length(gps.strains)
                strain = gps['strains'][istr] #strain = gps.strains{istr};
        
                source = i['Lf'][age][hiv]['v0'][strain]#source = i.Lf.(age).(hiv).v0.(strain);
                destin = i['Lf'][age][hiv]['v1'][strain] #destin = i.Lf.(age).(hiv).v1.(strain);
                rate   = r['vacc'][ia] #rate   = r.vacc(ia);
                m[destin,source] = m[destin,source] + rate #m(destin,source) = m(destin,source) + rate;
        
                source = i['Lf'][age][hiv]['v1'][strain] #source = i.Lf.(age).(hiv).v1.(strain);
                destin = i['Lf'][age][hiv]['v2'][strain] #destin = i.Lf.(age).(hiv).v2.(strain);
                rate   = r['waning'] #rate   = r.waning;
                m[destin,source] = m[destin,source] + rate #m(destin,source) = m(destin,source) + rate;
            #end
        #end
    #end

#% --- Introduction of Ageing rate                                           
    sources = s['a1'] #sources = s.a1;
    destins = s['a2'] #destins = s.a2;
    
    inds = np.ravel_multi_index((destins, sources), m.shape, order='C') ##inds = np.ravel_multi_index((destins, sources), m.shape, order='F') #inds    = sub2ind(size(m), destins, sources);
    #inds = np.ravel_multi_index([destins, sources], dims=np.shape(m), order='F')

    rate   = r['ageing'][0] #rate   = r.ageing(1);
    m.ravel()[inds] = m.ravel()[inds] + rate

    sources = s['a2'] #sources = s.a2;
    destins = s['a3'] #destins = s.a3;
    inds = np.ravel_multi_index((destins, sources), m.shape, order='C') ##inds = np.ravel_multi_index((destins, sources), m.shape, order='F') # inds    = sub2ind(size(m), destins, sources);
    rate   = r['ageing'][1] #rate   = r.ageing(2);
    m.ravel()[inds] = m.ravel()[inds] + rate #m(inds) = m(inds) + rate;

    sources = s['a3'] #sources = s.a3;
    destins = s['a4'] #destins = s.a4;
    inds = np.ravel_multi_index((destins, sources), m.shape, order='C') #inds = np.ravel_multi_index((destins, sources), m.shape, order='F') #inds    = sub2ind(size(m), destins, sources);
    rate   = r['ageing'][2] #rate   = r.ageing(3);
    m.ravel()[inds] = m.ravel()[inds] + rate #m(inds) = m(inds) + rate;

    sources = s['a4']#sources = s.a4;
    destins = s['a5'] #destins = s.a5;
    inds = np.ravel_multi_index((destins, sources), m.shape, order='C') ##inds = np.ravel_multi_index((destins, sources), m.shape, order='F') #inds    = sub2ind(size(m), destins, sources);
    rate   = r['ageing'][3] #rate   = r.ageing(4);
    m.ravel()[inds] = m.ravel()[inds] + rate #m(inds) = m(inds) + rate;

    #% --- HIV acquisition                                                      % <--- NEW
    m3 = np.zeros((i['nstates'],i['nstates'])) #m3 = zeros(i.nstates);
    sources = s['h0'] #sources = s.h0;
    destins = s['h1'] #destins = s.h1;
    inds = np.ravel_multi_index((destins, sources), m.shape, order='C') #inds = np.ravel_multi_index((destins, sources), m.shape, order='F') #inds    = sub2ind(size(m), destins, sources);
    rates   = 1 #rates   = 1;
    m3.ravel()[inds] = m3.ravel()[inds] + rates #   m3(inds) = m3(inds) + rates;

    #% --- ART initiation
    sources = s['h1'] #sources = s.h1;
    destins = s['hart'] #destins = s.hart;
    inds = np.ravel_multi_index((destins, sources), m.shape, order='C') #inds = np.ravel_multi_index((destins, sources), m.shape, order='F') ##inds = np.ravel_multi_index((destins, sources), m.shape, order='F') #inds    = sub2ind(size(m), destins, sources);
    rates   = r['ART_init'] #rates   = r.ART_init;
    m.ravel()[inds] = m.ravel()[inds] + rates #m(inds) = m(inds) + rates;

    #% --- Bring them together
    M = {}
    M['lin']   = sparse.csc_matrix(m - np.diag(np.sum(m,0))) #M.lin    = sparse(m - diag(sum(m,1)));
    M['Dxlin']  = sparse.csc_matrix(m2 - np.diag(np.sum(m2,0))) #M.Dxlin  = sparse(m2 - diag(sum(m2,1)));
    M['linHIV'] = sparse.csc_matrix(m3 - np.diag(np.sum(m3,0))) #M.linHIV = sparse(m3 - diag(sum(m3,1)));                                   % <--- NEW, now HIV acquisition is separate matrix made in lines 62-67
    M['linLTBI'] = sparse.csc_matrix(m4 - np.diag(np.sum(m4,0))) #M.linLTBI = sparse(m4 - diag(sum(m4,1)));
    #% --- Get the nonlinear rates ---------------------------------------------
    M['nlin'] = {}
    for ia in range(len(gps['age'])):
        age  = gps['age'][ia]
        M['nlin'][age] = {}

    #% --- Allocating transitions
    for istr in range(len(gps['strains'])): #istr = 1:length(gps.strains)
        strain = gps['strains'][istr]
      
        for ia in range(len(gps['age'])): #ia = 1:length(gps.age)
            age  = gps['age'][ia]
    
            m = np.zeros((i['nstates'],i['nstates'])) #m = zeros(i.nstates);
            for ih in range(len(gps['hiv'])) : #ih = 1:length(gps.hiv)
                hiv = gps['hiv'][ih] #hiv = gps.hiv{ih};
        
                for iv in range(len(gps['vacc'])): #iv = 1:length(gps.vacc)
                    vacc = gps['vacc'][iv] #vacc = gps.vacc{iv};
            
                    tidx0 = np.intersect1d(np.intersect1d(np.intersect1d(np.intersect1d(np.concatenate([s['Lf'], s['Ls'], s['Rlo'], s['Rhi'], s['R']]),s[age]),s[hiv]),s[strain]),s[vacc])
                    sources = [i['U'][age][hiv][vacc], tidx0] #sources = [i.U.(age).(hiv).(vacc), intersect(intersect(intersect(intersect([s.Lf, s.Ls, s.Rlo, s.Rhi, s.R],s.(age)),s.(hiv)),s.(strain)),s.(vacc))];
                    if len(tidx0) == 0:
                        sources = [i['U'][age][hiv][vacc]]
                    else:
                        sources = np.concatenate(sources)      
                    destin  = i['Lf'][age][hiv][vacc][strain] #destin  = i.Lf.(age).(hiv).(vacc).(strain);
                    m[destin, sources] = m[destin, sources] + (1-(iv==1)*p['VE'][0])#m(destin, sources) = m(destin, sources) + (1-(iv==2)*p.VE(1));
            
                    #%M.nlin.(hiv).(strain) = sparse(m - diag(sum(m,1)));
                #end
        
                #% Adjust for any immune protection
                cols = np.intersect1d(np.concatenate([s['Lf'], s['Ls'], s['Rlo'], s['Rhi'], s['R']]), s[hiv]) #cols = intersect([s.Lf, s.Ls, s.Rlo, s.Rhi, s.R], s.(hiv));
                m[:,cols] = m[:,cols]*(1-p['imm'][ih]) #m(:,cols) = m(:,cols)*(1-p.imm(ih));
            #end
            M['nlin'][age][strain] = sparse.csc_matrix(m - np.diag(np.sum(m,0))) #M.nlin.(age).(strain) = sparse(m - diag(sum(m,1)));
        #end
    #end

    #% construct basic building block of repeating contact matrix, as per five-age category
    tmp = np.zeros((len(gps['age']),i['nstates'])) #tmp = zeros(length(gps.age),i.nstates);
    for ia_sus in range(len(gps['age'])): #ia_sus = 1:length(gps.age)
        for ia_inf in range(len(gps['age'])): #ia_inf = 1:length(gps.age)
            cols = np.intersect1d(s['infectious'],s[gps['age'][ia_inf]]) #cols = intersect(s.infectious,s.(gps.age{ia_inf}));
            tmp[ia_sus,cols] = np.array(prm['cm'])[ia_sus,ia_inf] #tmp(ia_sus,cols) = prm.cm(ia_sus,ia_inf);
        #end
    #end    

    #% --- Getting force-of-infection
    m = np.zeros((10,i['nstates'])) #m = zeros(10,i.nstates);            % DS/mdr and 5 age-groups
    m[0:5,np.intersect1d(s['infectious'], s['ds'])]   = r['beta']*tmp[:, np.intersect1d(s['infectious'], s['ds'])] #m(1:5,intersect(s.infectious, s.ds))   = r.beta*tmp(:,intersect(s.infectious, s.ds));
    m[5:10,np.intersect1d(s['infectious'], s['mdr'])]  = r['beta']*p['rfbeta_mdr']*tmp[:,np.intersect1d(s['infectious'], s['mdr'])] #%r.beta*p.rfbeta_mdr #m(6:10,intersect(s.infectious, s.mdr))  = r.beta*p.rfbeta_mdr*tmp(:,intersect(s.infectious, s.mdr)); %r.beta*p.rfbeta_mdr;
    m[:,[s['h1'], s['hart']]] = m[:,[s['h1'], s['hart']]]*p['rfbeta_hiv'] #m(:,[s.h1, s.hart]) = m(:,[s.h1, s.hart])*p.rfbeta_hiv;
    m[:,s['Isc']] = m[:,s['Isc']]*0.75 #m(:,s.Isc) = m(:,s.Isc)*0.75;
    M['lambda'] = sparse.csc_matrix(m) #M.lambda = sparse(m);

 
    #% --- Get the mortality rates
    m = np.zeros((i['nstates'],5)) #m = zeros(i.nstates,5);
    #%m(:,1)            = r.mort;
    m[s['a1'],0]         = r['mort'][0] #m(s.a1,1)         = r.mort(1);
    m[s['a2'],0]         = r['mort'][1] #m(s.a2,1)         = r.mort(2);
    m[s['a3'],0]         = r['mort'][2] #m(s.a3,1)         = r.mort(3);
    m[s['a4'],0]         = r['mort'][3]#m(s.a4,1)         = r.mort(4);
    m[s['a5'],0]         = r['mort'][4] #m(s.a5,1)         = r.mort(5);

    m[s['h1'],0]         = r['HIV_mort'] #m(s.h1,1)         = r.HIV_mort;                                            % HIV+ve, no TB
    #% HIV-ve TB
    inds = np.intersect1d([s['h0']], s['infectious_wosc']) #inds = intersect([s.h0], s.infectious_wosc);
    m[inds,1] = r['mort_TB'][0] #m(inds,2) = r.mort_TB(1);
    #% HIV/TB coinfection
    inds = np.intersect1d([s['h1'],s['hart']], s['infectious_wosc']) #inds = intersect([s.h1,s.hart], s.infectious_wosc);
    m[inds,2] = r['mort_TB'][1] #m(inds,3) = r.mort_TB(2);
    #%keyboard

    M['mortvec'] = m #M.mortvec = m;

    return(M)


from SpectrumCommon.Const.BG import *
from SpectrumCommon.Modvars.GB.GBDefs import *


BG_modvar_shapes = {
        # bgmvt.BG_VERSION_TAG                              : bgc.BG_FILE_VERSION_1_0,
        # bgmvt.BG_BUDGET_RECORDS_TAG                       : budget_records,
        # bgmvt.BG_INDICATOR_RECORDS_IC_TAG                 : ind_records_IC,
        # bgmvt.BG_INDICATOR_RECORDS_PC_TAG                 : ind_records_PC,
        # bgmvt.BG_INDICATOR_RECORDS_IS_TAG                 : ind_records_IS,
        # bgmvt.BG_INDICATOR_RECORDS_HW_TAG                 : ind_records_HW,
        # bgmvt.BG_INDICATOR_RECORDS_HS_TAG                 : ind_records_HS,
        # bgmvt.BG_INDICATOR_RECORDS_GV_TAG                 : ind_records_GV,
        # bgmvt.BG_INDICATOR_RECORDS_HF_TAG                 : ind_records_HF,
        # bgmvt.BG_BUDGET_CAT_RECORDS_TAG                   : budget_cat_records,
        # bgmvt.BG_BUDGET_CAT_SHARE_RECORDS_IC_TAG          : [],
        # bgmvt.BG_BUDGET_CAT_SHARE_RECORDS_PC_TAG          : [],
        # bgmvt.BG_BUDGET_CAT_SHARE_RECORDS_IS_TAG          : [],
        # bgmvt.BG_BUDGET_CAT_SHARE_RECORDS_HW_TAG          : [],
        # bgmvt.BG_BUDGET_CAT_SHARE_RECORDS_HS_TAG          : [],
        # bgmvt.BG_BUDGET_CAT_SHARE_RECORDS_GV_TAG          : [],
        # bgmvt.BG_BUDGET_CAT_SHARE_RECORDS_HF_TAG          : [],
        # bgmvt.BG_FUND_FW_RECORDS_TAG                      : fund_framework_records,
        # bgmvt.BG_FUND_SOURCE_RECORDS_TAG                  : fund_source_records,
        # bgmvt.BG_FUND_SOURCE_SHARE_RECORDS_IC_TAG         : [],
        # bgmvt.BG_FUND_SOURCE_SHARE_RECORDS_PC_TAG         : [],
        # bgmvt.BG_FUND_SOURCE_SHARE_RECORDS_IS_TAG         : [],  
        # bgmvt.BG_FUND_SOURCE_SHARE_RECORDS_HW_TAG         : [],  
        # bgmvt.BG_FUND_SOURCE_SHARE_RECORDS_HS_TAG         : [],  
        # bgmvt.BG_FUND_SOURCE_SHARE_RECORDS_GV_TAG         : [],
        # bgmvt.BG_FUND_SOURCE_SHARE_RECORDS_HF_TAG         : [],  
        # bgmvt.BG_USING_SAME_FUND_MAPPINGS_ALL_BUDGETS_TAG : False,
        # bgmvt.BG_QUICK_MAP_RECORDS_TAG                    : [],

        # #icmvt.IC_DATA_CHANGED_TAG_V1                        : value_bool_False, 

        # #icmvt.IC_RES_NUM_SERVICES_BY_DELIV_CHAN_TAG_V1      : icmvd.IC_init_interv_deliv_chan_all_year_arr(num_intervs, num_deliv_chan_IC, num_years_outputs, 0.0).tolist(),
        
        # bgmvt.BG_RES_TOTAL_COSTS_BY_BUDGET_CAT_TAG             : [],
        # bgmvt.BG_RES_TOTAL_COSTS_BY_FUND_SOURCE_TAG            : [],
        # bgmvt.BG_RES_TOTAL_COSTS_BY_BUDGET_CAT_FUND_SOURCE_TAG : [],

        # bgmvt.BG_DEFAULT_BUDGET_LIST_META_TAG             : default_budget_list_meta,
        # bgmvt.BG_DEFAULT_FUND_FW_LIST_META_TAG            : default_fund_fw_list_meta,
}


from SpectrumCommon.Const.DP import DP_BigPopTag
from SpectrumCommon.Const.IC import IC_INTERV_RECORDS_TAG, IC_F_INTERV_ID
import SpectrumCommon.Const.NC as ncc
from SpectrumCommon.Const.PJ import PJN_FirstYearTag, PJN_FinalYearTag

def call_ncd_model(modvars):
    year_len = modvars[PJN_FinalYearTag]-[PJN_FirstYearTag]+1
    interv_id_list = []
    interv_info = []
    ncd_results = []
    for rec in modvars[IC_INTERV_RECORDS_TAG] :
        if(rec[IC_F_INTERV_ID] in interv_id_list):
            interv_info.append(rec)
            ncd_res = {
                IC_F_INTERV_ID:rec[IC_F_INTERV_ID],
                'targetPop': [0.0]*year_len,
            }
            ncd_results.append(ncd_res)

    # interv_coverages = [rec for rec in modvars[IC_INTERV_RECORDS_TAG] if rec['id']=='']
    
    package = {
        'firstYear':modvars[PJN_FirstYearTag], #Int
        'finalYear':modvars[PJN_FinalYearTag], #Int
        'interventions': interv_info, 
        'pop':modvars[DP_BigPopTag], #Sex, Age, Time
        'asthma':{
            'IncidenceResult'  : modvars[ncc.NC_AsthmaIncidenceResultTag],
            'PrevalnceResult'  : modvars[ncc.NC_AsthmaPrevalnceResultTag],
            'MortalityResult'  : modvars[ncc.NC_AsthmaMortalityResultTag],
            'RemissionResult'  : modvars[ncc.NC_AsthmaRemissionResultTag],
            'RecurrenceResult' : modvars[ncc.NC_AsthmaRecurrenceResultTag],
            'HYLResult'        : modvars[ncc.NC_AsthmaHYLResultTag],
            'YLDResult'        : modvars[ncc.NC_AsthmaYLDResultTag],
            'DWResult'         : modvars[ncc.NC_AsthmaDWResultTag],
        }
    }
    
    return None






# Notes:
#
#   1. You cannot base logic on whether a current MV tag exists in the code below.  It will always exist, because before this gets called, all missing MVs
#      will get added to the projection.
#
#   2. To trigger this, create a projection, say on the client, save, and grab it's projection ID.  Call GetModvars to get the library ID and put that into the appropriate
#      environment variable. Then call OpenProjection. Make sure to leave the projection on the client open while doing this because the proj ID changes when you open/close.
def LG_UpdateModvars(projection, created_projection):

    # if IS_FACILITY_MED_EQUIP_GROUP_RECORDS_TAG_V1 in projection:
    #     # String constants were added for facility medical equipment group records.
    #     facility_MEG_records = projection[IS_FACILITY_MED_EQUIP_GROUP_RECORDS_TAG_V1 ]

    #     for rec in facility_MEG_records:
    #         if not (ismvf.IS_FACIL_MEG_STR_CONST in rec):
    #             rec[ismvf.IS_FACIL_MEG_STR_CONST] = ''

    #     projection[ismvt.IS_FACILITY_MED_EQUIP_GROUP_RECORDS_TAG] = deepcopy(facility_MEG_records)

    # if IS_FACILITY_MED_EQUIP_RECORDS_TAG_V1 in projection:
    #     # String constants were added for facility equipment records.
    #     facility_equip_records = projection[IS_FACILITY_MED_EQUIP_RECORDS_TAG_V1]

    #     for rec in facility_equip_records:
    #         if not (ismvf.IS_FACIL_EQUIP_STR_CONST in rec):
    #             rec[ismvf.IS_FACIL_EQUIP_STR_CONST] = ''

    #     projection[ismvt.IS_FACILITY_MED_EQUIP_RECORDS_TAG] = deepcopy(facility_equip_records)

    # delete tags
    for tag in LG_TagsForDeletion:
        if tag in projection:
            del projection[tag]

    return projection

# IS_FACILITY_MED_EQUIP_RECORDS_TAG_V1           = '<IS_FacilityMedEquipRecords_V1>'
# IS_FACILITY_MED_EQUIP_GROUP_RECORDS_TAG_V1     = '<IS_FacilityMedEquipGroupRecords_V1>'

LG_TagsForDeletion = [
    # IS_FACILITY_MED_EQUIP_RECORDS_TAG_V1,
    # IS_FACILITY_MED_EQUIP_GROUP_RECORDS_TAG_V1
]



from os import environ as env
import numpy as np
import traceback

from AvenirCommon.Database import *
from AvenirCommon.Logger import *
from SpectrumCommon.Const.GB import GB_MOD_STR, GB_PJ, GB_SW_BADGETMODVARS, GB_SW_BADSETMODVARS
from SpectrumCommon.Const.PJ import PJN_ModulesTag
from .MakeAuthenticatedCalls import make_post
from SpectrumCommon.ServerUtils.ExceptionUtils import SpectrumException


def get_modvars(filter_packet):
    return make_post(env['AVENIR_SW_FILTER_URL'] +'/api/v1/getModvars', filter_packet)


def set_modvars(filter_packet):
    return make_post(env['AVENIR_SW_FILTER_URL'] +'/api/v1/setModvars', filter_packet)

def get_pjn_modvars(projID):
    return GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID+ '/PJ.JSON')

def get_all_modvars(projID):
    projection = GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID+ '/PJ.JSON')
    for mod_int in projection[PJN_ModulesTag]:
        if mod_int != GB_PJ:
            projection.update(GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID+ '/'+GB_MOD_STR[mod_int]+'.JSON'))

    _dtype = np.float64
    dt = np.dtype(np.float64)  
    for tag in projection:
        if type(projection[tag]) == list:
            try:
                if len(projection[tag])>0 and ((type(projection[tag][0])==dict) or(type(projection[tag][0])==str)) :
                    projection[tag] = np.array(projection[tag], order='C')#, dtype = dt)
                else:
                    projection[tag] = np.array(projection[tag], order='C', dtype=dt)#, dtype = dt)
            
            except:
                log('Failed to convert list to numpy array '+tag)
                projection[tag] = projection[tag] #nop
                 
    return projection

def get_calc_modvars(projID, mod_tags, modvars=None):
    try:
        _dtype = np.float64
        dt = np.dtype(np.float64)  
        if modvars==None:
            modvars = GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID+ '/PJ.JSON')
    
        packet = {
            "projID": projID,
            "modules": [{"module": GB_MOD_STR[mod_id], "modvars": mod_tags[mod_id] } for mod_id in modvars[PJN_ModulesTag] if (mod_id in mod_tags)]
            }
        response = get_modvars(packet)
        
        modules = ujson.loads(response.body)['result']
        for md in modules:
            for mv in md['modvars']:
                if ('value' in mv):
                    # print(mv['tag'])
                    if type(mv['value']) == list:
                        try:
                            if len(mv['value'])>0 and ((type(mv['value'][0])==dict) or(type(mv['value'][0])==str) or(type(mv['value'][0])==bool)) :
                                mv['value'] = np.array(mv['value'], order='C')#, dtype = dt)
                            else:
                                mv['value'] = np.array(mv['value'], order='C', dtype=dt)#, dtype = dt)
                        
                        except:
                            log('Failed to convert list to numpy array '+mv['tag'])
                            try: #try one more time
                                mv['value'] = np.array(mv['value'])#, dtype = dt)
                            except:
                                log('Failed Second  time to convert list to numpy array '+mv['tag'])
                                
                            mv['value'] = mv['value'] #nop

                        
                    
                    modvar = {mv['tag']:mv['value']}
                    modvars.update(modvar)
                
                
        # modvars = GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID + 'PJ.JSON')
        # modvars.update(GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID + 'AM.JSON'))
        # modvars.update(GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID + 'DP.JSON'))
        # modvars.update(GB_get_db_json(env['AVENIR_SW_STATE_DB_CONNECTION'], 'projections', projID + 'FP.JSON'))
        # for mv in modvars:
        #     #print(mv) 
        #     if mv in ['<AM_ChildMortByCD4WithART0to6_V1>', '<AM_ChildMortByCD4WithART0to6Perc_V1>',
        #               '<AM_ChildMortByCD4WithART7to12_V1>', '<AM_ChildMortByCD4WithART7to12Perc_V1>',
        #               '<AM_ChildMortByCD4WithARTGT12_V1>', '<AM_ChildMortByCD4WithARTGT12Perc_V1>']:
        #         #print('meow')
        #         continue
        #     if type(modvars[mv]) == list:
        #             if len(modvars[mv])>0 and ((type(modvars[mv][0])==dict) or(type(modvars[mv][0])==str)) :
        #                 modvars[mv] = np.array(modvars[mv], order='C')#, dtype = dt)
        #             else:
        #                 modvars[mv] = np.array(modvars[mv], order='C', dtype=dt)#, dtype = dt)
        return modvars
    except Exception as e:
        extra_info = {'msg':e.args[0] if len(e.args)>0 else''}
        tb = traceback.format_exception(e)
        logger.error(extra_info)
        logger.error(tb)
        raise SpectrumException(GB_SW_BADGETMODVARS, extra_info=extra_info) 
    
def set_all_modvars(projID, projection):
    return set_calc_modvars(projID, projection)

def set_calc_modvars(projID, projection):
    try:
        modules = {}

        numpy_type = type(np.array([]))

        for tag in projection:
            # log(tag)
            if tag[0] == '<':
                mod_str = tag[1:3]
                if mod_str not in modules:
                    modules[mod_str] = []
                    
                if type(projection[tag]) == numpy_type:
                    projection[tag] = np.nan_to_num(projection[tag], copy=True, nan=0.0, posinf=None, neginf=None)
                    projection[tag] = projection[tag].tolist()
                    
                #print(tag)
                modules[mod_str].append({'tag':tag, 'value':projection[tag]})
            
        
        packet = {
            "projID": projID,
            "modules":  [{'module':mod_str, 'modvars':modules[mod_str]} for mod_str in modules] 
        }
        pass
        response = set_modvars(packet)
            
        if response.status_code!=200:
            raise Exception('Setmodvars filter api call failed')
        
        ######## By  MODULE ####################
        # for mod_str in modules:
        #     packet = {
        #         "projID": projID,
        #         "modules": [{'module':mod_str, 'modvars':modules[mod_str]}]
        #     }
        
        #     response = set_modvars(packet)
            
        #     if response.status_code!=200:
        #         raise Exception('Setmodvars filter api call failed')

        
        # ######## By  MODVAR ####################
        # for mod_str in modules:
        #     for mv in modules[mod_str]:
        #         packet = {
        #             "projID": projID,
        #             "modules": [{'module':mod_str, 'modvars':mv}]
        #         }
        #         print(mv['tag'])
        #         # if mv['tag'] == '<FP_RelativeRisks_V1>':
        #         #     print(mv['value'])
        #         if mv['tag'] != '<FP_RelativeRisks_V1>':
        #             response = set_modvars(packet)
                    
        #             if response.status_code!=200:
        #                 raise Exception('Setmodvars filter api call failed')
    
    except Exception as e:
        extra_info = {'msg':e.args[0] if len(e.args)>0 else''}
        tb = traceback.format_exception(e)
        logger.error(extra_info)
        logger.error(tb)
        raise SpectrumException(GB_SW_BADSETMODVARS, extra_info=extra_info) 